/****************************************************************************

 ■HiResAmp アンプ制御(AmpControl基板)
		Copyright(C) 2020 電子うさぎ https://電子うさぎ.com/

****************************************************************************/

// PIC24F32KA304 Configuration Bit Settings

// FBS
#pragma config BWRP = OFF               // Boot Segment Write Protect (Disabled)
#pragma config BSS = OFF                // Boot segment Protect (No boot program flash segment)

// FGS
#pragma config GWRP = OFF               // General Segment Write Protect (General segment may be written)
#pragma config GSS0 = OFF               // General Segment Code Protect (No Protection)

// FOSCSEL
#pragma config FNOSC = FRC              // Oscillator Select (Fast RC Oscillator (FRC))
#pragma config SOSCSRC = DIG            // SOSC Source Type (Digital Mode for use with external source)
#pragma config LPRCSEL = HP             // LPRC Oscillator Power and Accuracy (High Power, High Accuracy Mode)
#pragma config IESO = ON                // Internal External Switch Over bit (Internal External Switchover mode enabled (Two-speed Start-up enabled))

// FOSC
#pragma config POSCMOD = NONE           // Primary Oscillator Configuration bits (Primary oscillator disabled)
#pragma config OSCIOFNC = OFF           // CLKO Enable Configuration bit (CLKO output disabled)
#pragma config POSCFREQ = HS            // Primary Oscillator Frequency Range Configuration bits (Primary oscillator/external clock input frequency greater than 8MHz)
#pragma config SOSCSEL = SOSCHP         // SOSC Power Selection Configuration bits (Secondary Oscillator configured for high-power operation)
#pragma config FCKSM = CSDCMD           // Clock Switching and Monitor Selection (Both Clock Switching and Fail-safe Clock Monitor are disabled)

// FWDT
#pragma config WDTPS = PS8192           // Watchdog Timer Postscale Select bits (1:8192)
#pragma config FWPSA = PR32             // WDT Prescaler bit (WDT prescaler ratio of 1:32)
#pragma config FWDTEN = SWON            // Watchdog Timer Enable bits (WDT controlled with the SWDTEN bit setting)
#pragma config WINDIS = OFF             // Windowed Watchdog Timer Disable bit (Standard WDT selected(windowed WDT disabled))

// FPOR
#pragma config BOREN = BOR3             // Brown-out Reset Enable bits (Brown-out Reset enabled in hardware, SBOREN bit disabled)
#pragma config LVRCFG = OFF             // Low Voltage Regulator Configuration bit (Low Voltage regulator is not available)
#pragma config PWRTEN = ON              // Power-up Timer Enable bit (PWRT enabled)
#pragma config I2C1SEL = SEC            // Alternate I2C1 Pin Mapping bit (Use  Alternate SCL1/ASDA1 Pins For I2C1)
#pragma config BORV = V27               // Brown-out Reset Voltage bits (Brown-out Reset set at 2.7V)
#pragma config MCLRE = ON               // MCLR Pin Enable bit (RA5 input pin disabled,MCLR pin enabled)

// FICD
#pragma config ICS = PGx1               // ICD Pin Placement Select bits (EMUC/EMUD share PGC1/PGD1)

// FDS
#pragma config DSWDTPS = DSWDTPSF       // Deep Sleep Watchdog Timer Postscale Select bits (1:2,147,483,648 (25.7 Days))
#pragma config DSBOREN = ON             // Deep Sleep Zero-Power BOR Enable bit (Deep Sleep BOR enabled in Deep Sleep)
#pragma config DSWDTEN = OFF            // Deep Sleep Watchdog Timer Enable bit (DSWDT disabled)

#include <xc.h>
#include <p24F32KA304.h>
#include <libpic30.h>

//=================================================================

#define I2C_SLAVE			0x51
#define I2C_INPUT			0x30
#define I2C_BIAS_LP			0x20
#define I2C_BIAS_LN			0x21
#define I2C_BIAS_RP			0x22
#define I2C_BIAS_RN			0x23
#define I2C_AK4499			0x11

#define DEBUG_LED			_LATA0
#define AVCC_EN				_LATA3
#define AK4499EQ_PDN		_LATA10

#define LMK04832_CS			_LATA7
#define LMK04832_SCK		_LATA6
#define LMK04832_SDIO		_RB10
#define LMK04832_SDIO_TRIS	_TRISB10
#define LMK04832_SDIO_CNPU	_CN16PUE
#define LMK04832_LD1		_RB11	// PLL1&PLL2 DLD
#define LMK04832_SEL0		_RB12	// LOS
#define VCXO_FS				_LATB13

#define BBST_ON				_LATB8
#define BBST_OFF			_LATB9
#define ATR1_A				_LATC6
#define ATR1_B				_LATC7
#define ATR2_A				_LATC8
#define ATR2_B				_LATC9
#define SP_RLY				_LATB4
#define HP_RLY				_LATA8
#define IN_MUTE				_RC2
#define IN_PLGD				_RC1

// 入力制御定義
#define COMBO384_F3_MASK	0x08
#define COMBO384_F2_MASK	0x04
#define COMBO384_F1_MASK	0x02
#define COMBO384_F0_MASK	0x01
#define COMBO384_F_MASK		0x0F
#define COMBO384_F_DSD64	(COMBO384_F3_MASK | COMBO384_F0_MASK)
#define COMBO384_F_DSD128	(COMBO384_F3_MASK | COMBO384_F1_MASK)
#define COMBO384_F_DSD256	(COMBO384_F3_MASK | COMBO384_F1_MASK | COMBO384_F0_MASK)
#define COMBO384_F_DSD512	(COMBO384_F3_MASK | COMBO384_F2_MASK)
#define COMBO384_RESERVED	0x80
#define COMBO384_DSDON_MASK	0x40
#define COMBO384_DSD64_MASK	0x20
#define COMBO384_MUTE_MASK	0x10

#define DSD_UNKNOWN			0
#define DSD_32KHZ			1		
#define DSD_44KHZ			2
#define DSD_48KHZ			3

// ステート値
#define STATE_INIT			0		// 初期化中(共通)
#define STATE_STANDBY		1		// 待機中
#define STATE_CONNECTED		2		// 接続済
#define STATE_PLAYING		3		// 再生中
#define STATE_ERROR			4		// エラー発生
#define STATE_SHUTDOWN		5		// シャットダウン(共通)

// イベント値
#define EVENT_NONE			0
#define EVENT_USB_CNTED		1		// USB接続
#define EVENT_INPUT_ON		2		// 信号入力ON
#define EVENT_INPUT_OFF		3		// 信号入力OFF
#define EVENT_USB_DCNT		4		// USB切断

// 再生フォーマット
#define AF_UNKNOWN			0
#define AF_I2S_1624			1
#define AF_I2S_32			2
#define AF_DSD_64			3
#define AF_DSD_128			4
#define AF_DSD_256			5
#define AF_DSD_512			6

// コマンド
#define CMD_SET_VOLUME		1		// ボリューム設定
#define CMD_SET_BBOOST		2		// バスブースト設定
#define CMD_SET_OUTPUT		3		// 出力切り替え(SP/HP)
#define CMD_SET_MUTE		4		// ミュート
#define CMD_SET_GAIN		5		// DACゲイン設定
#define CMD_SET_FLTPCM		6		// DACフィルタ設定(PCM)
#define CMD_SET_FLTDSD		7		// DACフィルタ設定(DSD)
#define CMD_SET_AUTOVL		8		// オートボリューム設定
#define CMD_SHUTDOWN		99		// シャットダウン(共通)

// 動作設定
#define AVC_VOL_LOWER		15		// オートボリューム設定可能範囲下限
#define AVC_VOL_UPPER		60		// オートボリューム設定可能範囲上限
#define AVC_CNT_LOWER		5		// オートボリューム可能範囲下限
#define AVC_CNT_UPPER		65		// オートボリューム可能範囲上限
#define AVC_SAMPLE			150		// サンプル数(200ms unit)
#define AVC_VOL_LIMIT		20		// ボリューム変更リミット±値
#define AVC_DISTIME			600		// 再生開始時ディセーブル時間(200ms unit)
#define AUTODN_INTERVAL		15		// ボリューム自動ダウン間隔(200ms unit)
#define AUTOUP_INTERVAL		100		// ボリューム自動アップ間隔(200ms unit)
#define AUTODN_MARGIN		5		// 超過検出マージン(%)
#define DLDCLR_INTERVAL		300		// LDクリア間隔(200ms unit)

//---------------------------------------------------------------

typedef enum _BOOL { FALSE = 0, TRUE } BOOL;
typedef unsigned char UINT8;
typedef unsigned short int UINT16;
typedef unsigned long int UINT32;

#pragma pack(push)
#pragma pack(1)
// 全体制御と共通↓↓↓↓↓
// LMK04832ステータス
typedef struct {
	UINT8 LOS			: 1;
	UINT8 PLL1_LD		: 1;
	UINT8 PLL2_LD		: 1;
	UINT8 PLL1_LD_LOST	: 1;
	UINT8 PLL2_LD_LOST	: 1;
	UINT8 HOLDOVER		: 1;
	UINT8 DAC_LOCKED	: 1;
	UINT16 DAC_VALUE;
} SLMK04832;

// バイアス制御ステータス
typedef struct {
	UINT8 nStatus;
	UINT8 nDACVal;
	UINT16 nAvgCur;
	UINT16 nSetCur;
} SBiasCtrl;

// ステータス総合
typedef struct {
	UINT8 bCmdBusy	: 1;	// コマンドBusy
	UINT8 nState	: 3;	// アンプ制御の状態
	UINT8 nErrInfo	: 4;	// エラー状態時の付加情報
	UINT8 nFormat	: 7;	// 再生フォーマット
	UINT8 nVCXO		: 1;	// VCXO周波数セレクト
	UINT8 nInputs;			// COMBO384 IO値
	UINT8 nInputInfo;		// 量子化ビット数又はDDSサンプリング周波数
	UINT8 nAnaATT;			// アナログアッテネータ
	UINT8 nDigATT;			// デジタルアッテネータ
	UINT8 bAdvValid;		// 音声平均電圧有効
	UINT16 nAdvAvg;			// 音声平均電圧
	SLMK04832 LMK04832;
	SBiasCtrl BiasCtrl[4];
} SStatus;

// 動作設定
typedef struct {
	UINT8 bEnable;			// 0or1
	UINT16 nVolume;			// 音量設定時Volume
	UINT16 nAdcVal;			// 音量設定値(ADC値)	
} SAutoVolCtrl;

typedef struct {
	UINT8 nVolume[2];		// HP=0～72, SP=0～81
	UINT8 nBBoost[2];		// 0or1
	UINT8 nOutput	: 1;	// 0:HP, 1:SP
	UINT8 nMute		: 1;	// 0or1
	UINT8 nPcmFlt	: 3;	// 0～6(4と5は同じ)
	UINT8 nDsdFlt	: 1;	// 0or1
	UINT8 nDacGain[2];		// 0or1
	SAutoVolCtrl Avc[2];
} SSetting;

// 送信データ
typedef union {
	struct {
		SStatus Status;
		SSetting Setting;
	};
	UINT8 Bytes[sizeof(SStatus) + sizeof(SSetting)];
} SAmpCtrl;
// 全体制御と共通↑↑↑↑↑
#pragma pack(pop)

//---------------------------------------------------------------

static void InitSystem(void);
static void InitApp(void);
static void Run(void);
static void ShutDown(void);

static void GetStatus(void);
static int GetEvent(void);
static void SetMCLKFreq(int nFreqNo);
static void SetVolume(int nVolume, BOOL bFirst, BOOL bAvc);
static void SetBBoost(BOOL bEnable);
static void SetOutput(int nType);
static void SetMute(BOOL bMute);
static void SetDacGain(int nGain);
static void SetFilterPCM(int nMode);
static void SetFilterDSD(int nMode);
static void SetAutoVolume(BOOL bEnable);
static void ClearAdvAvg(void);

static void RELAY_SetAtt(int nLevel);
static void RELAY_SetBB(BOOL bEnable);
static void RELAY_SetOutput(int nKind);

static BOOL AK4499EQ_Init(void);
static BOOL AK4499EQ_Check(void);
static void AK4499EQ_Reset(void);
static void AK4499EQ_PowerDown(void);
static void AK4499EQ_Operation(int nFormat, int nDsdSam);
static void AK4499EQ_SetMute(BOOL bMute);
static void AK4499EQ_SetVolume(UINT8 nATT);
static void AK4499EQ_SetGain(UINT8 nGain);

static BOOL LMK04832_Init(void);
static BOOL LMK04832_Check(void);
static void LMK04832_PowerDown(void);
static void LMK04832_Program(const UINT32* pRegs);
static void LMK04832_GetStatus(SLMK04832* pStatus);
static UINT8 LMK04832_ReadSPI(UINT16 addr);
static void LMK04832_WriteSPI(UINT32 addrdata);

static void SendI2C(UINT8 nAddr, const UINT8* pData, int nCount);
static void RecvI2C(UINT8 nAddr, UINT8* pData, int nCount);
static void RecvI2CEx(UINT8 nAddr, UINT8 nSub, UINT8* pData, int nCount);
static void WriteEEPROM(UINT16 nOffset, UINT16 nData);
static UINT16 ReadEEPROM(UINT16 nOffset);
static void DelayUs(UINT32 us);

#define DelayMs(x)	DelayUs((x) * 1000UL)

//---------------------------------------------------------------

const UINT32 s_LMK04832Regs1[];
const UINT32 s_LMK04832Regs2[];

static volatile SSetting s_Setting = { { 0 }, { 0 }, 0, 0, 0, 0, { 0 }, { { 0 }, { 0 } } };
static volatile SStatus s_Status = { 0 };
static volatile UINT16 s_ADV[AVC_SAMPLE] = { 0 };
static volatile UINT32 s_nADVTotal;
static volatile int s_nAdvCount;
static volatile int s_nAvcExec;
static volatile int s_nLD1Clear;
static volatile int s_nLD2Clear;
static volatile BOOL s_bUpdate;
static volatile UINT8 s_Recv[2];

//=================================================================

// ●メインエントリ
int main(void)
{
	InitSystem();
	InitApp();
	Run();
	ShutDown();
	return 0;
}

// ●システム初期化
static void InitSystem(void)
{
	// ポート初期化
	PORTA = PORTB = 0x00;
	LATA = LATB =0x00;
	
	// デバッグLED
	TRISAbits.TRISA0 = 0;
	ANSELAbits.ANSA0 = 0;
	DEBUG_LED = 1;
	
	// AVCC_EN
	TRISAbits.TRISA3 = 0;
	ANSELAbits.ANSA3 = 0;

	// AK4499EQ_PDN
	TRISAbits.TRISA10 = 0;

	// LMK04832_CS
	_LATA7 = 1;
	_TRISA7 = 0;
	
	// LMK04832_SCK
	_TRISA6 = 0;
	
	// LMK04832_SDIO
	_TRISB10 = 0;
	
	// LMK04832_LD1
	_TRISB11 = 1;
	
	// LMK04832_SEL0
	_TRISB12 = 1;
	ANSELBbits.ANSB12 = 0;

	// VCXO_FS
	_TRISB13 = 0;
	ANSELBbits.ANSB13 = 0;
	
	 // ピークホールド(C116)ディスチャージ開始
	TRISAbits.TRISA2 = 0;
	
	// BBST_ON
	TRISBbits.TRISB8 = 0;

	// BBST_OFF
	TRISBbits.TRISB9 = 0;

	// ATR1_A
	TRISCbits.TRISC6 = 0;

	// ATR1_B
	TRISCbits.TRISC7 = 0;

	// ATR2_A
	TRISCbits.TRISC8 = 0;

	// ATR2_B
	TRISCbits.TRISC9 = 0;

	// SP_RLY
	TRISBbits.TRISB4 = 0;
	ANSELBbits.ANSB4 = 0;

	// HP_RLY
	TRISAbits.TRISA8 = 0;

	// IN_MUTE
	TRISCbits.TRISC2 = 1;
	ANSELCbits.ANSC2 = 0;

	// IN_PLGD
	TRISCbits.TRISC1 = 1;
	ANSELCbits.ANSC1 = 0;
	
	// ADC
	AD1CON1bits.SSRC = 0b0010; // Timer3 event ends sampling and starts conversion
	AD1CON1bits.ASAM = 1; // Sampling begins immediately after the last conversion; SAMP bit is auto-set
	AD1CON1bits.MODE12 = 1;
	AD1CON2bits.PVCFG =  0b10; // 2 * Internal VBG(1.024V))
	AD1CON3bits.ADRC = 1; // RC clock
	AD1CHSbits.CH0NA = 0b000; //AVSS
	AD1CHSbits.CH0SA = 0b01101; // AN13
	_AD1IP = 3;
	_AD1IE = 1;

	// I2C1(AC1 Slave)
	I2C1CONbits.STREN = 1;
	I2C1ADD = I2C_SLAVE;
	_SI2C1IP = 4;
	_SI2C1IE = 1;

	// I2C2(AC2 Master)
	ANSELBbits.ANSB2 = 0;
	ANSELBbits.ANSB3 = 0;
	I2C2BRG = 4; // 400KHz at FCY=4MHz
	I2C2CONbits.DISSLW = 1; // ノイズ低減＆アイソレータにより固まりやすくなるため
	I2C2CONbits.I2CEN = 1;
	
	// Timer1(ATT LatchRelay)
	T1CONbits.TCS = 0; // Internal clock (FOSC/2)
	T1CONbits.TCKPS = 0b00; // 1:1 (0.25us, FCY=4MHz)
	PR1 = 40000; // 10ms後にラッチリレーOFF
	_T1IP = 6; // 優先度高い
	_T1IE = 1;
	
	// Timer2(Volume save)
	T2CONbits.TCS = 0; // Internal clock (FCY)
	T2CONbits.TCKPS = 0b11; // 1:256 (64us, FCY=4MHz)
	PR2 = (2000000UL / 64); // 2秒後に保存
	_T2IP = 2; // 優先度低い
	_T2IE = 1;

	// Timer3(SP/HP Relay → IntervalTimer(ADC Trigger))
	T3CONbits.TCS = 0; // Internal clock (FCY)
	T3CONbits.TCKPS = 0b11; // 1:256 (64us, FCY=4MHz)
	PR3 = (1000000UL / 64); // 1秒後にリレーON
	_T3IP = 2; // 優先度低い
	_T3IE = 1;
	
	// Timer4/5(Delay)
	T4CONbits.T32 = 1;
	T4CONbits.TCS = 0; // Internal clock (FCY)
	T4CONbits.TCKPS = 0b00; // 1:1 (0.25us, FCY=4MHz)
}

// ●アプリ初期化
static void InitApp(void)
{
	s_nAvcExec = 0;
	s_nLD1Clear = 0;
	s_nLD2Clear = 0;
	s_bUpdate = TRUE;

	// EEPROMから動作設定を復元する
	UINT16 nVal = ReadEEPROM(0); // 0:マーカー値
	if (nVal == 0x7777)
	{
		s_Setting.nVolume[0] = (UINT8)ReadEEPROM(2);
		s_Setting.nVolume[1] = (UINT8)ReadEEPROM(4);
		s_Setting.nBBoost[0] = (UINT8)ReadEEPROM(6);
		s_Setting.nBBoost[1] = (UINT8)ReadEEPROM(8);
		s_Setting.nOutput = (UINT8)ReadEEPROM(10);
		s_Setting.nPcmFlt = (UINT8)ReadEEPROM(12);
		s_Setting.nDsdFlt = (UINT8)ReadEEPROM(14);
		s_Setting.nDacGain[0] = (UINT8)ReadEEPROM(16);
		s_Setting.nDacGain[1] = (UINT8)ReadEEPROM(18);
		s_Setting.Avc[0].bEnable = (UINT8)ReadEEPROM(20);
		s_Setting.Avc[0].nVolume = (UINT8)ReadEEPROM(22);
		s_Setting.Avc[0].nAdcVal = (UINT8)ReadEEPROM(24);
		s_Setting.Avc[1].bEnable = (UINT8)ReadEEPROM(26);
		s_Setting.Avc[1].nVolume = (UINT8)ReadEEPROM(28);
		s_Setting.Avc[1].nAdcVal = (UINT8)ReadEEPROM(30);
	}
	else
	{
		// 初期状態→デフォルト値を設定して書き込んでおく
		s_Setting.nVolume[0] = 15;
		s_Setting.nVolume[1] = 10;
		s_Setting.nBBoost[0] = 0;
		s_Setting.nBBoost[1] = 1;
		s_Setting.nOutput = 0;
		s_Setting.nPcmFlt = 2;
		s_Setting.nDsdFlt = 0;
		s_Setting.nDacGain[0] = 1;
		s_Setting.nDacGain[1] = 0;
		s_Setting.Avc[0].bEnable = 0;
		s_Setting.Avc[0].nVolume = 0;
		s_Setting.Avc[0].nAdcVal = 0;
		s_Setting.Avc[1].bEnable = 0;
		s_Setting.Avc[1].nVolume = 0;
		s_Setting.Avc[1].nAdcVal = 0;
		
		WriteEEPROM(2, s_Setting.nVolume[0]);
		WriteEEPROM(4, s_Setting.nVolume[1]);
		WriteEEPROM(6, s_Setting.nBBoost[0]);
		WriteEEPROM(8, s_Setting.nBBoost[1]);
		WriteEEPROM(10, s_Setting.nOutput);
		WriteEEPROM(12, s_Setting.nPcmFlt);
		WriteEEPROM(14, s_Setting.nDsdFlt);
		WriteEEPROM(16, s_Setting.nDacGain[0]);
		WriteEEPROM(18, s_Setting.nDacGain[1]);
		WriteEEPROM(20, s_Setting.Avc[0].bEnable);
		WriteEEPROM(22, s_Setting.Avc[0].nVolume);
		WriteEEPROM(24, s_Setting.Avc[0].nAdcVal);
		WriteEEPROM(26, s_Setting.Avc[1].bEnable);
		WriteEEPROM(28, s_Setting.Avc[1].nVolume);
		WriteEEPROM(30, s_Setting.Avc[1].nAdcVal);

		WriteEEPROM(0, 0x7777);
	}
	// 保存されていない設定
	s_Setting.nMute = 0;
	
	// I2C受付開始
	I2C1CONbits.I2CEN = 1;
		
	// LMK04832初期化
	if (!LMK04832_Init())
	{
		s_Status.nState = STATE_ERROR;
		s_Status.nErrInfo |= 0x01; // Bit0:LMK04832エラー
		DEBUG_LED = 0;
	}

	// AVCCの電源を入れて安定待ち
	AVCC_EN = 1;
	DelayMs(2000); // AVCCのソフトスタート時定数を考慮

	// DAC初期化
	if (!AK4499EQ_Init())
	{
		s_Status.nState = STATE_ERROR;
		s_Status.nErrInfo |= 0x02; // Bit1:AK4499EQエラー
		DEBUG_LED = 0;
	}
	
	if (s_Status.nState != STATE_ERROR)
	{
		// Volume
		SetVolume(s_Setting.nVolume[s_Setting.nOutput], TRUE, FALSE);

		// BassBoost
		RELAY_SetBB(s_Setting.nBBoost[s_Setting.nOutput]);

		// 出力リレーON(1秒後)
		T3CONbits.TON = 1;

		// WDTタイムアウト検知のため
		if (RCONbits.WDTO)
			s_Status.nErrInfo |= 0x04; // Bit2:WDTタイムアウト検知

		// 通常動作開始
		s_Status.nState = STATE_STANDBY;
	}
	
	// 音量検知初期化
	ClearAdvAvg();
}

// ●実行ループ
static void Run(void)
{
#ifndef _DEBUG
	_SWDTEN = 1;
#endif
	// シャットダウンまで無限ループ
	for(;;)
	{
		 ClrWdt();
		 
		// 各デバイスの情報を取得する
		if (s_bUpdate)
		{
			// I2C通信で固まる可能性があるため正常時のみ
			if (s_Status.nState != STATE_ERROR)
				GetStatus();
			
			s_bUpdate = FALSE;
		}		

		// ステートマシン
		switch(GetEvent())
		{
		case EVENT_USB_CNTED: // USB接続
			if (s_Status.nState == STATE_STANDBY)
			{
				s_Status.nState = STATE_CONNECTED;
			}
			else if (s_Status.nState == STATE_CONNECTED)
			{
				 // ありえないはず
				s_Status.nState = STATE_ERROR;
				s_Status.nErrInfo |= 0x08; // Bit8:状態遷移エラー
				DEBUG_LED = 0;
			}
			break;

		case EVENT_INPUT_ON: // 信号入力ON
			if (s_Status.nState == STATE_PLAYING)
			{
				// DACミュート&リセット
				AK4499EQ_Reset();
			}
			if (s_Status.nState == STATE_STANDBY || s_Status.nState == STATE_CONNECTED || s_Status.nState == STATE_PLAYING)
			{
				// ジッタクリーナーのセットアップ
				// No1: 44.1KHz, 88.2KHz, 176.4KHz, 352.8KHz → MCLK:22.5792MHz
				// No2: 32KHz, 48KHz, 96KHz, 192KHz, 384KHz → MCLK:24.5760MHz
				int nFreqNo;
				if (s_Status.nFormat <= AF_I2S_32)
				{
					// PCMの場合
					nFreqNo = (s_Status.nInputs & COMBO384_F0_MASK)? 1 : 2;
				}
				else
				{
					// DSDの場合
					nFreqNo = 1;
					if ((s_Status.nInputInfo == DSD_32KHZ) || (s_Status.nInputInfo == DSD_48KHZ))
						nFreqNo = 2;
				}
				SetMCLKFreq(nFreqNo);

				// DACのセットアップ
				AK4499EQ_Operation(s_Status.nFormat, s_Status.nInputInfo);

				// DACミュート解除
				AK4499EQ_SetMute(s_Setting.nMute);
				
				s_Status.nState = STATE_PLAYING;
			}
			break;

		case EVENT_INPUT_OFF: // 信号入力OFF
			if (s_Status.nState == STATE_PLAYING)
			{
				// DACミュート&リセット
				AK4499EQ_Reset();

				s_Status.nState = STATE_CONNECTED;
			}
			else if (s_Status.nState == STATE_CONNECTED)
			{
				 // ありえないはず
				s_Status.nState = STATE_ERROR;
				s_Status.nErrInfo |= 0x08; // Bit8:状態遷移エラー
				DEBUG_LED = 0;
			}
			break;

		case EVENT_USB_DCNT: // USB切断
			if (s_Status.nState == STATE_PLAYING)
			{
				// DACミュート&リセット
				AK4499EQ_Reset();
				
				// 平均値を取り直す
				ClearAdvAvg();

				s_Status.nState = STATE_STANDBY;
			}
			else if (s_Status.nState == STATE_CONNECTED)
			{
				s_Status.nState = STATE_STANDBY;
			}
			else if (s_Status.nState == STATE_STANDBY)
			{
				 // ありえないはず
				s_Status.nState = STATE_ERROR;
				s_Status.nErrInfo |= 0x08; // Bit8:状態遷移エラー
				DEBUG_LED = 0;
			}
			break;
		}
		
		// コマンド実行
		if (s_Status.bCmdBusy)
		{
			int nCmd = s_Recv[0];
			int nArg = s_Recv[1];
			int nVal;

			if (nCmd == CMD_SHUTDOWN)
			{
				// シャットダウン →Busyのまま抜ける
				s_Status.nState = STATE_SHUTDOWN;
				return;
			}
			
			// 以下は正常時のみ
			if (s_Status.nState != STATE_ERROR)	
			{
				switch(nCmd)
				{
				case CMD_SET_VOLUME: // ボリューム設定
					nVal = (nArg)? 1 : -1; // 1step
					if (s_Setting.nOutput == 1)
					{
						// SPの場合少ないVolue値でのStep数を上げる
						if (nVal > 0)
						{
							// 上げる場合
							if (s_Setting.nVolume[1] <= 15)
								nVal *= 3;
							else if (s_Setting.nVolume[1] <= 28)
								nVal *= 2;
						}
						else
						{
							// 下げる場合
							if (s_Setting.nVolume[1] <= 18)
								nVal *= 3;
							else if (s_Setting.nVolume[1] <= 30)
								nVal *= 2;
						}
					}
					SetVolume(s_Setting.nVolume[s_Setting.nOutput] + nVal, FALSE, FALSE);
					break;
				case CMD_SET_BBOOST: // バスブースト設定
					SetBBoost(nArg);
					break;
				case CMD_SET_OUTPUT: // 出力切り替え(SP/HP)
					SetOutput(nArg);
					break;
				case CMD_SET_MUTE: // ミュート
					SetMute(nArg);
					break;
				case CMD_SET_GAIN: // DACゲイン設定
					SetDacGain(nArg);
					break;
				case CMD_SET_FLTPCM: // DACフィルタ設定(PCM)
					SetFilterPCM(nArg);
					break;
				case CMD_SET_FLTDSD: // DACフィルタ設定(DSD)
					SetFilterDSD(nArg);
					break;
				case CMD_SET_AUTOVL: // オートボリューム設定
					SetAutoVolume(nArg);
					break;
				}
			}
			s_Status.bCmdBusy = FALSE;
		}
		
		// オートボリュームを実行
		if (s_nAvcExec != 0)
		{
			// 正常時のみ
			if (s_Status.nState != STATE_ERROR)
			{
				int nCurrent = s_Setting.nVolume[s_Setting.nOutput];
				if (AVC_CNT_LOWER < nCurrent && nCurrent < AVC_CNT_UPPER)
					SetVolume(nCurrent + s_nAvcExec, FALSE, TRUE);
			}

			s_nAvcExec = 0;
		}
	}
}

// ●シャットダウン
static void ShutDown(void)
{
	const UINT8 addr[4] = { I2C_BIAS_LP, I2C_BIAS_LN, I2C_BIAS_RP, I2C_BIAS_RN };

	if (!(s_Status.nErrInfo & 0x02)) // Bit1:AK4499EQエラー
	{
		// DACミュート&リセット
		AK4499EQ_Reset();
	}
	
	// 出力リレーOFF
	T3CONbits.TON = 0;
	RELAY_SetOutput(-1); // SP/HPともOFF

	// 割り込み要因停止
	T2CONbits.TON = 0;
	T3CONbits.TON = 0;
	AD1CON1bits.ADON = 0;

	// AVCCの電源を切る
	AVCC_EN = 0;

	// バイアス制御へシャットダウンコマンド
	UINT8 i, data = CMD_SHUTDOWN;
	for(i = 0; i < 4; i++)
		SendI2C(addr[i], &data, 1);

	// LMK04832パワーダウン
	LMK04832_PowerDown();

	// AK4499EQパワーダウン
	AK4499EQ_PowerDown();

	// ピークホールド(C116)ディスチャージ
	TRISAbits.TRISA2 = 0;
	
	// AVCC抜き時間
	DelayMs(1000);

	// バイアス制御終了待ち
	for(i = 0; i < 4; i++)
	{
		do {
			RecvI2C(addr[i], &data, 1);
			DelayMs(10);
		}
		while(data != STATE_SHUTDOWN);
	}
	
	// 電源断まで無限ループ
	s_Status.bCmdBusy = FALSE;
	for(;;) ClrWdt();
}

// ●各デバイスの情報を取得する
static void GetStatus(void)
{
	// 入力監視からの情報
	// 接続直後のチャタリング除去と安定待ち
	static int s_nPLGD = 0;
	do {
		int nTemp = IN_PLGD;
		if (nTemp != s_nPLGD)
		{
			s_nPLGD = nTemp;
			if (s_nPLGD)
				DelayMs(10);
		}
	} while (s_nPLGD != IN_PLGD);

	if (s_nPLGD && !IN_MUTE)
	{
		// 信号を入力している→情報を受信
		UINT8 data[2];
		RecvI2C(I2C_INPUT, data, 2);
		s_Status.nInputs = data[0]; 
		s_Status.nInputInfo = data[1];

		// オーディオフォーマットを決定する
		if (!(s_Status.nInputs & COMBO384_RESERVED))
		{
			if (!(s_Status.nInputs & COMBO384_DSDON_MASK))
			{
				// I2Sの場合
				switch (s_Status.nInputInfo) // 量子化ビット数
				{
				case 16: case 24:
					s_Status.nFormat = AF_I2S_1624;
					break;
				case 32:
					s_Status.nFormat = AF_I2S_32;
					break;
				case 0:
					// ビット数が検出された後に判断中(0)になった場合は
					// ミュートでない限り変化させない
					if (IN_MUTE || (s_Status.nInputs & COMBO384_MUTE_MASK))
					{
						s_Status.nFormat = AF_UNKNOWN;
					}
					else if (s_Status.nFormat == AF_UNKNOWN)
					{
						// 有効なサンプル周波数ならとりあえず32bitにしておく
						int val = s_Status.nInputs & COMBO384_F_MASK;
						if (0x01 <= val && val <= 0x08) // 44.1kHz～384kHz
							s_Status.nFormat = AF_I2S_32;
					}
					break;
				}
			}
			else
			{
				// DSDの場合
				switch (s_Status.nInputs & COMBO384_F_MASK)
				{
				case COMBO384_F_DSD64:
					s_Status.nFormat = AF_DSD_64;
					break;
				case COMBO384_F_DSD128:
					s_Status.nFormat = AF_DSD_128;
					break;
				case COMBO384_F_DSD256:
					s_Status.nFormat = AF_DSD_256;
					break;
				case COMBO384_F_DSD512:
					s_Status.nFormat = AF_DSD_512;
					break;
				default:
					s_Status.nFormat = AF_UNKNOWN;
					break;
				}
			}
		}
		else
		{
			// データ異常(I2C通信異常など)
			s_Status.nInputs = 0;
			s_Status.nInputInfo = 0;
			s_Status.nFormat = AF_UNKNOWN;
		}
	}
	else
	{
		s_Status.nInputs = 0;
		s_Status.nInputInfo = 0;
		s_Status.nFormat = AF_UNKNOWN;
	}
	
	// 以下はコマンドを受信していない場合のみ(コマンド処理を優先する)
	if (!s_Status.bCmdBusy)
	{
		// VCXO
		s_Status.nVCXO = VCXO_FS;

		// LMK04832
		SLMK04832 temp;
		LMK04832_GetStatus(&temp);
		s_Status.LMK04832 = temp;

		// バイアス制御
		if (s_Status.bCmdBusy)
			return;
		RecvI2C(I2C_BIAS_LP, (UINT8*)&s_Status.BiasCtrl[0], 6);
		if (s_Status.bCmdBusy)
			return;
		RecvI2C(I2C_BIAS_LN, (UINT8*)&s_Status.BiasCtrl[1], 6);
		if (s_Status.bCmdBusy)
			return;
		RecvI2C(I2C_BIAS_RP, (UINT8*)&s_Status.BiasCtrl[2], 6);
		if (s_Status.bCmdBusy)
			return;
		RecvI2C(I2C_BIAS_RN, (UINT8*)&s_Status.BiasCtrl[3], 6);
	}
}

// ●イベントを取得する
static int GetEvent(void)
{
	static int s_nPLGD = 0;
	static int s_nFormat = AF_UNKNOWN;
	
	// USB 接続/切断
	int nTemp = IN_PLGD;
	if (nTemp != s_nPLGD)
	{
		s_nPLGD = nTemp;
		return (nTemp)? EVENT_USB_CNTED : EVENT_USB_DCNT;
	}

	// 信号入力 ON/OFF
	if (s_Status.nFormat != s_nFormat)
	{
		s_nFormat = s_Status.nFormat;
		return (s_nFormat == AF_UNKNOWN)? EVENT_INPUT_OFF : EVENT_INPUT_ON;
	}
	
	return EVENT_NONE;
}

// ●MCLK周波数を設定する
static void SetMCLKFreq(int nFreqNo)
{
	static int s_LastFreq = -1;
	
	if (nFreqNo != s_LastFreq)
	{
		if (nFreqNo == 1)
		{
			// 22.5792MHz
			VCXO_FS = 0;

			// Settling Time after FS Change : 10ms
			DelayMs(10);

			LMK04832_Program(s_LMK04832Regs1);
		}
		else
		{
			// 24.57602MHz
			VCXO_FS = 1;

			// Settling Time after FS Change : 10ms
			DelayMs(10);

			LMK04832_Program(s_LMK04832Regs2);
		}

		s_LastFreq = nFreqNo;
	}
}

// ●ボリュームを設定する
static void SetVolume(int nVolume, BOOL bFirst, BOOL bAvc)
{
	if (s_Setting.nOutput == 0)
	{
		// HPの場合
		if (nVolume > 62) nVolume = 62;
		else if (nVolume < 0) nVolume = 0;
	}
	else
	{
		// SPの場合
		if (nVolume > 72) nVolume = 72;
		else if (nVolume < 0) nVolume = 0;
	}

	int nCurrent = s_Setting.nVolume[s_Setting.nOutput];
	if (!bFirst && (nVolume == nCurrent))
		return;

	BOOL bEnlarge = bFirst || (nVolume > nCurrent);

	// 先にアッテネータ設定
	UINT8 nAtt = 0;
	if (bEnlarge)
	{
		if (nVolume <= 48)
			nAtt = 2; // -24dB
		else if (nVolume <= 60)
			nAtt = 1; // -12dB
	}
	else
	{
		if (nVolume <= 46)
			nAtt = 2; // -24dB
		else if (nVolume <= 58)
			nAtt = 1; // -12dB
	}
	if (bFirst || (nAtt != s_Status.nAnaATT))
	{
		RELAY_SetAtt(nAtt); // タイマーにより10ms後にラッチOFF
		s_Status.nAnaATT = nAtt;
	}
	
	// 次にDAC設定
	UINT8 nDac = 0;
	if (nVolume > 0)
	{
		int nMax;
		if (bEnlarge)
		{
			if (nVolume <= 48) nMax = 48;
			else if (nVolume <= 60) nMax = 60;
			else nMax = 72;
		}
		else
		{
			if (nVolume <= 46) nMax = 48;
			else if (nVolume <= 58) nMax = 60;
			else nMax = 72;
		}
		
		nDac = 0xFF - ((nMax - nVolume) * 2);
		if (nVolume < 16)
		{
			// 減衰率1.5倍
			nDac -= 16 - nVolume; // -0.5dB
		}
		if (nVolume < 6)
		{
			// 減衰率2倍
			nDac -= 6 - nVolume; // -0.5dB
		}
	}
	
	AK4499EQ_SetVolume(nDac);
	s_Status.nDigATT = nDac;

	if (!bFirst)
	{
		s_Setting.nVolume[s_Setting.nOutput] = nVolume;
		
		if (!bAvc)
		{
			// 平均値を取り直す
			ClearAdvAvg();

			// オートボリューム設定値をリセット
			if (s_Setting.Avc[s_Setting.nOutput].bEnable)
			{
				// ボリュームが設定可能範囲外では設定できない
				if (nVolume < AVC_VOL_LOWER || nVolume > AVC_VOL_UPPER)
					SetAutoVolume(FALSE);
				else
				{
					_AD1IE = 0;
					s_Setting.Avc[s_Setting.nOutput].nVolume = 0;
					s_Setting.Avc[s_Setting.nOutput].nAdcVal = 0;
					_AD1IE = 1;
				}
			}
		}
		// 2秒後に保存を行う
		T2CONbits.TON = 0;
		TMR2 = 0;
		T2CONbits.TON = 1;
	}
}

// ●バスブーストを設定する
static void SetBBoost(BOOL bEnable)
{
	s_Setting.nBBoost[s_Setting.nOutput] = bEnable;

	if (s_Setting.nOutput == 0)
		WriteEEPROM(6, bEnable);
	else
		WriteEEPROM(8, bEnable);

	RELAY_SetBB(bEnable);
}

// ●出力を設定する
static void SetOutput(int nType)
{
	s_Setting.nOutput = nType;
	WriteEEPROM(10, nType);

	// 一旦完全OFF
	RELAY_SetOutput(-1);

	// SP⇔HPゲイン切り替え
	AK4499EQ_SetGain(s_Setting.nDacGain[nType]);
	
	// SP⇔HPバスブースト切り替え
	RELAY_SetBB(s_Setting.nBBoost[nType]);

	// SP⇔HPボリューム切り替え
	SetVolume(s_Setting.nVolume[nType], TRUE, FALSE);
	
	// 新しい出力先へ
	RELAY_SetOutput(nType);

	// 平均値を取り直す
	ClearAdvAvg();
}

// ●ミュート
static void SetMute(BOOL bMute)
{
	s_Setting.nMute = bMute;
	AK4499EQ_SetMute(bMute);
}

// ●DACゲインの設定(PCM)
static void SetDacGain(int nGain)
{
	if (s_Setting.nOutput == 0)
	{
		s_Setting.nDacGain[0] = nGain;
		WriteEEPROM(16, nGain);
	}
	else
	{
		s_Setting.nDacGain[1] = nGain;
		WriteEEPROM(18, nGain);
	}
	
	AK4499EQ_SetGain(nGain);
	
	// 平均値を取り直す
	ClearAdvAvg();

	// オートボリューム設定値をリセット
	if (s_Setting.Avc[s_Setting.nOutput].bEnable)
	{
		_AD1IE = 1;
		s_Setting.Avc[s_Setting.nOutput].nVolume = 0;
		s_Setting.Avc[s_Setting.nOutput].nAdcVal = 0;
		_AD1IE = 1;

		if (s_Setting.nOutput == 0)
		{
			WriteEEPROM(22, s_Setting.Avc[0].nVolume);
			WriteEEPROM(24, s_Setting.Avc[0].nAdcVal);
		}
		else
		{
			WriteEEPROM(28, s_Setting.Avc[1].nVolume);
			WriteEEPROM(30, s_Setting.Avc[1].nAdcVal);
		}
	}
}

// ●PCMデジタルフィルタの設定(PCM)
static void SetFilterPCM(int nMode)
{
	s_Setting.nPcmFlt = nMode;
	WriteEEPROM(12, s_Setting.nPcmFlt);
	
	if (AF_I2S_1624 == s_Status.nFormat || s_Status.nFormat == AF_I2S_32)
	{
		// 一旦リセット(&ミュート)して設定する
		AK4499EQ_Reset();
		AK4499EQ_Operation(s_Status.nFormat, 0);
		AK4499EQ_SetMute(s_Setting.nMute);
	}
}

// ●DSDカットオフ周波数の設定(DDS)
static void SetFilterDSD(int nMode)
{
	s_Setting.nDsdFlt = nMode;
	WriteEEPROM(14, s_Setting.nDsdFlt);

	if (AF_DSD_64 <= s_Status.nFormat && s_Status.nFormat <= AF_DSD_512)
	{
		// 一旦リセット(&ミュート)して設定する
		AK4499EQ_Reset();
		AK4499EQ_Operation(s_Status.nFormat, s_Status.nInputInfo);
		AK4499EQ_SetMute(s_Setting.nMute);
	}
}

// ●オートボリューム設定
static void SetAutoVolume(BOOL bEnable)
{
	volatile SAutoVolCtrl* pAvc = &s_Setting.Avc[s_Setting.nOutput];
	
	if (pAvc->bEnable == bEnable)
		return;
	
	int nSave = _AD1IE;

	if (bEnable)
	{
		// ボリュームが設定可能範囲外では設定できない
		int nVolume = s_Setting.nVolume[s_Setting.nOutput];
		if (nVolume < AVC_VOL_LOWER || nVolume > AVC_VOL_UPPER)
		{
			_AD1IE = nSave;
			return;
		}
		
		// 設定
		pAvc->bEnable = TRUE;
		if (s_Status.bAdvValid)
		{
			// すぐにオートボリュームを設定する
			pAvc->nVolume = s_Setting.nVolume[s_Setting.nOutput];
			pAvc->nAdcVal = s_Status.nAdvAvg;

		}
		else
		{
			// 平均値が有効になった時に設定する
			pAvc->nVolume = 0;
			pAvc->nAdcVal = 0;
		}
	}
	else
	{
		// 解除
		pAvc->bEnable = FALSE;
		pAvc->nVolume = 0;
		pAvc->nAdcVal = 0;
	}
	
	_AD1IE = nSave;	

	// 保存
	if (s_Setting.nOutput == 0)
	{
		WriteEEPROM(20, s_Setting.Avc[0].bEnable);
		WriteEEPROM(22, s_Setting.Avc[0].nVolume);
		WriteEEPROM(24, s_Setting.Avc[0].nAdcVal);
	}
	else
	{
		WriteEEPROM(26, s_Setting.Avc[1].bEnable);
		WriteEEPROM(28, s_Setting.Avc[1].nVolume);
		WriteEEPROM(30, s_Setting.Avc[1].nAdcVal);
	}
}

// ●平均値をリセットする
static void ClearAdvAvg(void)
{
	 // ピークホールド(C116)ディスチャージ開始
	TRISAbits.TRISA2 = 0;

	int nSave = _AD1IE;
	_AD1IE = 0;

	// 平均値を取り直す
	int i = 0;
	do s_ADV[i] = 0;
	while (++i < AVC_SAMPLE);

	s_nAdvCount = 0;
	s_nADVTotal = 0;
	
	s_Status.bAdvValid = FALSE;
	s_Status.nAdvAvg = 0;

	_AD1IE = nSave;	

	 // ピークホールド(C116)ディスチャージ終了
	DelayMs(5);
	TRISAbits.TRISA2 = 1;
}

//---------------------------------------------------------------

// ●アッテネータ設定
static void RELAY_SetAtt(int nLevel)
{
	while(T1CONbits.TON);

	switch(nLevel)
	{
	case 0: // -0dB
		ATR1_B = 1;
		ATR2_A = 1;
		break;
	case 1: // ATT1(-12dB)
		ATR1_A = 1;
		ATR2_B = 1;
		break;
	case 2: // ATT2(-24dB)
		ATR1_B = 1;
		ATR2_B = 1;
		break;
	}
	// 10ms後にOFFする
	TMR1 = 0;
	T1CONbits.TON = 1;
}

// ●バスブースト設定
static void RELAY_SetBB(BOOL bEnable)
{
	if (bEnable)
	{
		BBST_ON = 1;
		DelayMs(10);
		BBST_ON = 0;
	}
	else
	{
		BBST_OFF = 1;
		DelayMs(10);
		BBST_OFF = 0;
	}
}

// ●出力リレー
static void RELAY_SetOutput(int nKind)
{
	if (nKind == 1)
	{
		// SP
		HP_RLY = 0;
		SP_RLY = 1;
	}
	else if (nKind == 0)
	{
		// HP
		SP_RLY = 0;
		HP_RLY = 1;
	}
	else
	{
		// Off
		SP_RLY = 0;
		HP_RLY = 0;
	}
}

//---------------------------------------------------------------

// ●AK4499初期化
static BOOL AK4499EQ_Init(void)
{
	UINT8 data[2];

	AK4499EQ_PDN = 1;
	DelayMs(2); // > 2ms
	
	// パワーアップ直後のレジスタを見る
	if (!AK4499EQ_Check())
		return FALSE;

	// Control2
	UINT8 SMUTE = 1;
	UINT8 DEM1 = 0b01;
	UINT8 SD = 1;

	data[0] = 0x01;
	data[1] = (SMUTE << 0) | (DEM1 << 1) | (SD << 5);
	SendI2C(I2C_AK4499, data, 2);
	
	// Control5
	UINT8 SYNCE = 0;
	
	data[0] = 0x07;
	data[1] = (SYNCE << 0);
	SendI2C(I2C_AK4499, data, 2);
	
	// Control6
	UINT8 DEM2 = 0b01;
	UINT8 PW1 = 1;
	UINT8 PW2 = 1;

	data[0] = 0x0A;
	data[1] = (DEM2 << 0) | (PW1 << 2) | (PW2 << 3);
	SendI2C(I2C_AK4499, data, 2);

	// Control7
	UINT8 MONO2 = 1;
	
	data[0] = 0x0B;
	data[1] = (MONO2 << 5);
	SendI2C(I2C_AK4499, data, 2);
	
	return TRUE;
}

// ●AK4499の正常動作を確認する(パワーアップ直後のみ有効)
static BOOL AK4499EQ_Check(void)
{
	// レジスタのデフォルト値をチェックする
	UINT8 data[4];
	RecvI2CEx(I2C_AK4499, 0, data, 4);
	
	// Control1
	if (data[0] != 0x0C)
		return FALSE;
	
	// Control2
	if (data[1] != 0x22)
		return FALSE;
	
	// Control3
	if (data[2] != 0x00)
		return FALSE;

	// L1ch ATT
	if (data[3] != 0xFF)
		return FALSE;
	
	return TRUE;
}

// ●DACをミュートしリセット(スタンバイ)状態にする
static void AK4499EQ_Reset(void)
{
	UINT8 data[2];

	// Control2
	data[0] = 0x01;
	RecvI2CEx(I2C_AK4499, data[0], &data[1], 1);
	if (!(data[1] & 0x01))
	{
		data[1] |= 0x01; // SMUTEを付加する
		SendI2C(I2C_AK4499, data, 2);

		// Figure 60. Soft Mute Function
		DelayMs(100);  // transition time
	}

	// Control1
	data[0] = 0x00;
	data[1] = 0x0C; // Control1のデフォルト値
	SendI2C(I2C_AK4499, data, 2);
	
	DelayMs(100); // 3～4/fs
}

// ●AK4499パワーダウン
static void AK4499EQ_PowerDown(void)
{
	AK4499EQ_PDN = 0;
}

// ●DACを通常動作状態にする
static void AK4499EQ_Operation(int nFormat, int nDsdSam)
{
	UINT8 data[2];
	
	// Control1(RSTN以外)
	UINT8 DIF = (nFormat != AF_I2S_32)? 0b011 : 0b111;
	UINT8 AFSD = 1;
	
	data[0] = 0x00;
	data[1] = (DIF << 1) | (AFSD << 4);
	SendI2C(I2C_AK4499, data, 2);
	
	// Control2
	UINT8 SD = (s_Setting.nPcmFlt & 0x2)? 1 : 0;
	UINT8 SMUTE = s_Setting.nMute;

	data[0] = 0x01;
	RecvI2CEx(I2C_AK4499, data[0], &data[1], 1);
	data[1] = (data[1] & ~0x20) | (SD << 5) | (SMUTE << 0);
	SendI2C(I2C_AK4499, data, 2);
	
	// Control3
	UINT8 SLOW = s_Setting.nPcmFlt & 0x1;
	UINT8 SELLR1 = (nFormat < AF_DSD_64)? 1 : 0;
	UINT8 MONO1 = 1;
	UINT8 DCKS = (nFormat >= AF_DSD_64 && nDsdSam == DSD_32KHZ)? 1 : 0;
	UINT8 DP = (nFormat == AF_I2S_1624 || nFormat == AF_I2S_32)? 0 : 1;

	data[0] = 0x02;
	data[1] = (SLOW << 0) | (SELLR1 << 1) | (MONO1 << 3) | (DCKS << 5) | (DP << 7);
	SendI2C(I2C_AK4499, data, 2);

	// Control4
	UINT8 SELLR2 = (nFormat < AF_DSD_64)? 0 : 1;
	UINT8 SSLOW = (s_Setting.nPcmFlt & 0x4)? 1 : 0;

	data[0] = 0x05;
	RecvI2CEx(I2C_AK4499, data[0], &data[1], 1);
	data[1] = (data[1] & ~0x09) | (SELLR2 << 3) | (SSLOW << 0);
	SendI2C(I2C_AK4499, data, 2);

	// Control5
	UINT8 GC = (s_Setting.nDacGain[s_Setting.nOutput] == 0)? 0b00 : 0b11;

	data[0] = 0x07;
	RecvI2CEx(I2C_AK4499, data[0], &data[1], 1);
	data[1] = (data[1] & ~0x06) | (GC << 1);
	SendI2C(I2C_AK4499, data, 2);

	// DSD1
	UINT8 DSDSEL0 = 0;
	if (nFormat == AF_DSD_128) DSDSEL0 = 1;
	if (nFormat == AF_DSD_512) DSDSEL0 = 1;
	UINT8 DDM = 1;

	data[0] = 0x06;
	data[1] = (DSDSEL0 << 0) | (DDM << 7);
	SendI2C(I2C_AK4499, data, 2);

	// DSD2
	UINT8 DSDF = s_Setting.nDsdFlt;
	UINT8 DSDSEL1 = 0;
	if (nFormat == AF_DSD_256) DSDSEL1 = 1;
	if (nFormat == AF_DSD_512) DSDSEL1 = 1;
	UINT8 DSDPATH = 1;

	data[0] = 0x09;
	data[1] = (DSDSEL1 << 0) | (DSDF << 1) | (DSDPATH << 2);
	SendI2C(I2C_AK4499, data, 2);

	// Control1(RSTN)
	data[0] = 0x00;
	RecvI2CEx(I2C_AK4499, data[0], &data[1], 1);
	data[1] |= 0x01;
	SendI2C(I2C_AK4499, data, 2);
}

// ●AK4499のMUTE設定
static void AK4499EQ_SetMute(BOOL bMute)
{
	UINT8 data[2];
	
	// Control2
	data[0] = 0x01;
	RecvI2CEx(I2C_AK4499, data[0], &data[1], 1);
	
	if (bMute)
	{
		if (!(data[1] & 0x01))
		{
			data[1] |= 0x01; // SMUTEを付加する
			SendI2C(I2C_AK4499, data, 2);

			// Figure 60. Soft Mute Function
			DelayMs(100);  // transition time
		}
	}
	else
	{
		if ((data[1] & 0x01))
		{
			data[1] &= ~0x01; // SMUTEを除去する
			SendI2C(I2C_AK4499, data, 2);

			// Figure 60. Soft Mute Function
			DelayMs(100);  // transition time
		}
	}
}

// ●AK4499のデジタルボリューム設定
static void AK4499EQ_SetVolume(UINT8 nATT)
{
	UINT8 data[2];
	data[1] = nATT;

	// L1ch ATT
	data[0] = 0x03;
	SendI2C(I2C_AK4499, data, 2);

	// R1ch ATT 
	data[0] = 0x04;
	SendI2C(I2C_AK4499, data, 2);

	// L2ch ATT
	data[0] = 0x0C;
	SendI2C(I2C_AK4499, data, 2);

	// R2ch ATT
	data[0] = 0x0D;
	SendI2C(I2C_AK4499, data, 2);
	
	DelayMs(100);  // transition time
}

// ●AK4499のゲイン設定
static void AK4499EQ_SetGain(UINT8 nGain)
{
	UINT8 data[2];
	
	// Control5
	data[0] = 0x07;
	RecvI2CEx(I2C_AK4499, data[0], &data[1], 1);
	
	UINT8 GC = (nGain == 0)? 0b00 : 0b11; // 72.8mApp : 45.5mApp
	data[1] = (data[1] & 0xF9) | (GC << 1);
	SendI2C(I2C_AK4499, data, 2);
}

//---------------------------------------------------------------

// ●LMK04832初期化
static BOOL LMK04832_Init(void)
{
	// Reset
	LMK04832_WriteSPI(0x000080); // 3 Wire Mode enabled
	DelayMs(1); // 念の為

	return LMK04832_Check();
}

// ●各IDをチェックして正常動作を確認する
static BOOL LMK04832_Check(void)
{
	UINT8 data;
			
	// ID_DEVICE_TYPE
	data = LMK04832_ReadSPI(0x0003);
	if (data != 6)
		return FALSE;
	
	// ID_PROD[15:8]
	data = LMK04832_ReadSPI(0x0004);
	if (data != 99) // 209は恐らくデータシート誤り
		return FALSE;

	// ID_PROD[7:0]
	data = LMK04832_ReadSPI(0x0005);
	if (data != 209) // 99は恐らくデータシート誤り
		return FALSE;
	
	// ID_MASKREV
	data = LMK04832_ReadSPI(0x0006);
	if (data != 112)
		return FALSE;

	// ID_VNDR[15:8]
	data = LMK04832_ReadSPI(0x000C);
	if (data != 81)
		return FALSE;

	// ID_VNDR[7:0]
	data = LMK04832_ReadSPI(0x000D);
	if (data != 4)
		return FALSE;

	return TRUE;
}

// ●LMK04832パワーダウン
static void LMK04832_PowerDown(void)
{
	LMK04832_WriteSPI(0x000201);
}

// ●全レジスタ設定
static void LMK04832_Program(const UINT32* pRegs)
{
	do {
		LMK04832_WriteSPI(*pRegs);
		pRegs++;
	}
	while (*pRegs != 0xFFFFFF);
}

// ●ステータスを返す
static void LMK04832_GetStatus(SLMK04832* pStatus)
{
	static UINT8 s_nLastLD1 = 0;
	static UINT8 s_nLastLD2 = 0;
	
	UINT8 data;
	
	// DLD関連
	data = LMK04832_ReadSPI(0x183);
	pStatus->PLL1_LD_LOST = (data & 0x08) != 0;
	pStatus->PLL1_LD = (data & 0x04) != 0;
	pStatus->PLL2_LD_LOST = (data & 0x02) != 0;
	pStatus->PLL2_LD = (data & 0x01) != 0;
	
	// ロックを検知したらクリアタイマーをリセットする
	if (pStatus->PLL1_LD != s_nLastLD1)
	{
		s_nLastLD1 = pStatus->PLL1_LD;
		if (s_nLastLD1)
			s_nLD1Clear = 0;
	}
	if (pStatus->PLL2_LD != s_nLastLD2)
	{
		s_nLastLD2 = pStatus->PLL2_LD;
		if (s_nLastLD2)
			s_nLD2Clear = 0;
	}
	// LD_LOSTクリア処理
	UINT8 flags = 0;
	if (s_nLD1Clear >= DLDCLR_INTERVAL)
	{
		s_nLD1Clear = 0;
		if (pStatus->PLL1_LD && pStatus->PLL1_LD_LOST)
			flags |= 0x02; // CLR_PLL1_LD_LOST
	}
	if (s_nLD2Clear >= DLDCLR_INTERVAL)
	{
		s_nLD2Clear = 0;
		if (pStatus->PLL2_LD && pStatus->PLL2_LD_LOST)
			flags |= 0x01; // CLR_PLL2_LD_LOST
	}
	if (flags != 0)
	{
		LMK04832_WriteSPI(0x018200 | flags);
		LMK04832_WriteSPI(0x018200);
	}
	
	// LOS
	data = LMK04832_ReadSPI(0x184);
	pStatus->LOS = (data & 0x01) != 0; // RB_CLKin0_LOS
	
	// HOLDOVER
	data = LMK04832_ReadSPI(0x188);
	pStatus->HOLDOVER = (data & 0x20) != 0; // RB_HOLDOVER
	pStatus->DAC_LOCKED = (data & 0x01) != 0; // RB_DAC_LOCKED

	// RB_DAC_VALUE
	data = LMK04832_ReadSPI(0x184) << 8;
	pStatus->DAC_VALUE = data | LMK04832_ReadSPI(0x185);
}

// ●LMK04832 3WIRE式SPIリード ※2MHzクロック&最適化レベル2以下を想定
static UINT8 LMK04832_ReadSPI(UINT16 addr)
{
	UINT8 data = 0;
	int i;
	
	LMK04832_SCK = 0;
	LMK04832_CS = 0;
	
	// R/W:READ(1)
	LMK04832_SDIO = 1;
	LMK04832_SCK = 1;

	// A14～A0
	for(i = 14; i >= 0; i--)
	{
		LMK04832_SCK = 0;
		LMK04832_SDIO = ((addr >> i) & 1);
		LMK04832_SCK = 1;
	}
	
	// ポートをSDIOからの入力へ
	LMK04832_SDIO_TRIS = 1;
	LMK04832_SDIO_CNPU = 1; // SDIO_RDBK_TYPがデフォルト(open drain)でも読めるようにする
	
	// D7～D0
	for(i = 7; i >= 0; i--)
	{
		LMK04832_SCK = 0;
		Nop(); Nop(); // max. 60ns
		data |= (UINT8)(LMK04832_SDIO << i);
		LMK04832_SCK = 1;
	}

	LMK04832_CS = 1;
	LMK04832_SCK = 0;
	LMK04832_SDIO = 0;
	
	// 出力に戻しておく
	LMK04832_SDIO_CNPU = 0;
	LMK04832_SDIO_TRIS = 0;

	return data;
}

// ●LMK04832 3WIRE式SPIライト
static void LMK04832_WriteSPI(UINT32 addrdata)
{
	int i;
	
	LMK04832_SCK = 0;
	LMK04832_CS = 0;
	
	// R/W:WRITE(0)
	LMK04832_SDIO = 0;
	LMK04832_SCK = 1;

	// A14～A0, D7～D0
	for(i = 22; i >= 0; i--)
	{
		LMK04832_SCK = 0;
		LMK04832_SDIO = ((addrdata >> i) & 1);
		LMK04832_SCK = 1;
	}

	LMK04832_CS = 1;
	LMK04832_SCK = 0;
	LMK04832_SDIO = 0;
}

//---------------------------------------------------------------

// ●I2C送信
static void SendI2C(UINT8 nAddr, const UINT8* pData, int nCount)
{
	// Start conditon
	I2C2CONbits.SEN = 1;
	while (I2C2CONbits.SEN);

	// 7bit address & W flag
	I2C2TRN = (nAddr << 1) | 0x00;
	while (I2C2STATbits.TRSTAT);

	// data
	int i;
	for(i = 0; i < nCount; i++)
	{
		I2C2TRN = pData[i];
		while (I2C2STATbits.TRSTAT);
	}

	// Stop condition
	I2C2CONbits.PEN = 1;
	while (I2C2CONbits.PEN);
}

// ●I2C受信
static void RecvI2C(UINT8 nAddr, UINT8* pData, int nCount)
{
	// Start conditon
	I2C2CONbits.SEN = 1;
	while (I2C2CONbits.SEN);

	// 7bit address & R flag
	I2C2TRN = (nAddr << 1) | 0x01;
	while (I2C2STATbits.TRSTAT);

	// data
	int i = 0;
	do {
		// data
		I2C2CONbits.RCEN = 1;
		while(I2C2CONbits.RCEN);
		pData[i++] = I2C2RCV;

		// ACK or NAK
		I2C2CONbits.ACKDT = (i == nCount)? 1 : 0;
		I2C2CONbits.ACKEN = 1;
		while(I2C2CONbits.ACKEN);

	} while(!I2C2CONbits.ACKDT);

	// Stop condition
	I2C2CONbits.PEN = 1;
	while (I2C2CONbits.PEN);
}

// ●I2C受信
static void RecvI2CEx(UINT8 nAddr, UINT8 nSub, UINT8* pData, int nCount)
{
	// Start conditon
	I2C2CONbits.SEN = 1;
	while (I2C2CONbits.SEN);

	// 7bit address & W flag
	I2C2TRN = (nAddr << 1) | 0x00;
	while (I2C2STATbits.TRSTAT);

	// sub
	I2C2TRN = nSub;
	while (I2C2STATbits.TRSTAT);

	// Start conditon
	I2C2CONbits.SEN = 1;
	while (I2C2CONbits.SEN);

	// 7bit address & R flag
	I2C2TRN = (nAddr << 1) | 0x01;
	while (I2C2STATbits.TRSTAT);

	// data
	int i = 0;
	do {
		// data
		I2C2CONbits.RCEN = 1;
		while(I2C2CONbits.RCEN);
		pData[i++] = I2C2RCV;

		// ACK or NAK
		I2C2CONbits.ACKDT = (i == nCount)? 1 : 0;
		I2C2CONbits.ACKEN = 1;
		while(I2C2CONbits.ACKEN);

	} while(!I2C2CONbits.ACKDT);

	// Stop condition
	I2C2CONbits.PEN = 1;
	while (I2C2CONbits.PEN);
}

// ●EEPROM(DATA)へ書き込み
static void WriteEEPROM(UINT16 nOffset, UINT16 nData)
{
	// PIC24FV32KA304 Family DATA EEPROM address 0x7FFE00～0x7FFFFF(256Word)
	_prog_addressT addr = 0x7FFE00 + nOffset;

	_erase_eedata(addr, _EE_WORD);
	_wait_eedata();

	_write_eedata_word(addr, nData);
	_wait_eedata();
}

// ●EEPROM(DATA)へ読み出し
static UINT16 ReadEEPROM(UINT16 nOffset)
{
	// PIC24FV32KA304 Family DATA EEPROM address 0x7FFE00～0x7FFFFF(256Word)
	_prog_addressT addr = 0x7FFE00 + nOffset;

	UINT16 nData;
	_memcpy_p2d16(&nData, addr, _EE_WORD);
	return nData;
}

// ●USウェイト(min.15us, max.1073.741823s)
static void DelayUs(UINT32 us)
{
	us = us << 2; // 0.25us → 1us

	PR4 = (UINT16)us;
	PR5 = (UINT16)(us >> 16);
	TMR4 = 0;
	TMR5 = 0;

	_T5IF = 0;
	T4CONbits.TON = 1;
	while(!_T5IF);
	T4CONbits.TON = 0;
}

//---------------------------------------------------------------

// ●Timer1割り込み
void __attribute__((interrupt, auto_psv)) _T1Interrupt(void)
{
	// アッテネータ用ラッチリレーをOFFにする
	ATR1_A = 0;
	ATR1_B = 0;
	ATR2_A = 0;
	ATR2_B = 0;

	T1CONbits.TON = 0;
	_T1IF = 0;
}

// ●Timer2割り込み
void __attribute__((interrupt, auto_psv)) _T2Interrupt(void)
{
	// ボリューム値、オートボリューム設定値(=0)保存
	if (s_Setting.nOutput == 0)
	{
		WriteEEPROM(2, s_Setting.nVolume[0]);
		WriteEEPROM(22, s_Setting.Avc[0].nVolume);
		WriteEEPROM(24, s_Setting.Avc[0].nAdcVal);
	}
	else
	{
		WriteEEPROM(4, s_Setting.nVolume[1]);
		WriteEEPROM(28, s_Setting.Avc[1].nVolume);
		WriteEEPROM(30, s_Setting.Avc[1].nAdcVal);
	}

	T2CONbits.TON = 0;
	_T2IF = 0;
}

// ●Timer3割り込み
void __attribute__((interrupt, auto_psv)) _T3Interrupt(void)
{
	// 最初の一発目 →出力リレーON
	RELAY_SetOutput(s_Setting.nOutput);

	// オートボリュームコントロール用のADCトリガーに切り替える
	T3CONbits.TON = 0;
	PR3 = (200000UL / 64); // 200ms
	TMR3 = 0;
	_T3IE = 0;

	AD1CON1bits.ADON = 1;
	AD1CON1bits.SAMP = 1;
	T3CONbits.TON = 1;

	_T3IF = 0;
}

// ●ADC割り込み(Timer3による定期トリガーのため汎用インターバルタイマ(200ms)としても利用する)
void __attribute__((interrupt, auto_psv)) _ADC1Interrupt(void)
{
	static volatile UINT32 s_nPlayTime = 0;
	static volatile int s_nAutoUp = 0;
	static volatile int s_nAutoDn = 0;
	static volatile int s_nAutoDnCnt = 0;

	// 再生時間
	if (s_Status.nFormat == AF_UNKNOWN)
	{
		// 0にリセットするのではなく5倍のペースで減らしていく
		if (s_nPlayTime > 5)
			s_nPlayTime -= 5;
		else
			s_nPlayTime = 0;
	}
	else
	{
		++s_nPlayTime;
	}
	
	if (s_nPlayTime > AVC_DISTIME)
	{
		volatile SAutoVolCtrl* pAvc = &s_Setting.Avc[s_Setting.nOutput];

		// 過去 AVC_SAMPLE x200ms 分の平均をとる
		UINT16 nOld = s_ADV[s_nAdvCount];
		UINT16 nNew = ADC1BUF0;
		s_ADV[s_nAdvCount] = nNew;

		s_nADVTotal -= nOld;
		s_nADVTotal += nNew;
		s_Status.nAdvAvg = s_nADVTotal / AVC_SAMPLE;

		if (++s_nAdvCount >= AVC_SAMPLE)
		{
			s_nAdvCount = 0;

			// 平均値が有効になった
			s_Status.bAdvValid = TRUE;

			if (pAvc->bEnable && pAvc->nAdcVal == 0)
			{
				// オートボリュームを設定する
				pAvc->nVolume = s_Setting.nVolume[s_Setting.nOutput];
				pAvc->nAdcVal = s_Status.nAdvAvg;

				if (s_Setting.nOutput == 0)
				{
					WriteEEPROM(22, s_Setting.Avc[0].nVolume);
					WriteEEPROM(24, s_Setting.Avc[0].nAdcVal);
				}
				else
				{
					WriteEEPROM(28, s_Setting.Avc[1].nVolume);
					WriteEEPROM(30, s_Setting.Avc[1].nAdcVal);
				}
			}
		}

		// オートボリューム設定が有効な場合
		if (s_Status.bAdvValid && pAvc->bEnable && pAvc->nAdcVal != 0)
		{
			long nLeft = (long)s_Status.nAdvAvg * 100;
			long nRight = (long)pAvc->nAdcVal * (100 + AUTODN_MARGIN);
			if (nLeft >= nRight)
			{
				// 超過を検出した
				if (s_nAutoDn == 0)
				{
					int nDiff = s_Setting.nVolume[s_Setting.nOutput] - pAvc->nVolume;
					if (nDiff < 0) nDiff *= 1;
					if (nDiff < AVC_VOL_LIMIT)
						s_nAvcExec = -1;

					if (s_nAutoDnCnt < 5)
						s_nAutoDnCnt++;
								
					s_nAutoDn = AUTODN_INTERVAL * s_nAutoDnCnt + 1;
				}
				s_nAutoUp = 0;
			}
			else if (s_Status.nAdvAvg < pAvc->nAdcVal)
			{
				// 未満になっている →所定のペースでボリュームを上げる
				if (++s_nAutoUp >= AUTOUP_INTERVAL)
				{
					// 音量アップ(戻す)
					int nDiff = s_Setting.nVolume[s_Setting.nOutput] - pAvc->nVolume;
					if (nDiff < 0) nDiff *= 1;
					if (nDiff < AVC_VOL_LIMIT)
						s_nAvcExec = 1;

					s_nAutoUp = 0;
					s_nAutoDn = 0;
					s_nAutoDnCnt = 0;
				}
			}
			else if (s_Status.nAdvAvg == pAvc->nAdcVal)
			{
				// 全く同じ場合
				if (s_nAutoUp > 0)
					--s_nAutoUp;
			}
			if (s_nAutoDn > 0)
				--s_nAutoDn;
		}
		else
		{
			s_nAutoUp = 0;
			s_nAutoDn = 0;
			s_nAutoDnCnt = 0;
		}
	}
	else
	{
		s_nAutoUp = 0;
		s_nAutoDn = 0;
		s_nAutoDnCnt = 0;
	}
	
	// CLR_PLL_LD_LOSTクリアのタイミングを作る
	s_nLD1Clear++;
	s_nLD2Clear++;
	
	// ステータス更新のタイミングを作る
	s_bUpdate = TRUE;

	_AD1IF = 0;
}

// ●I2C1スレーブ割り込み
void __attribute__((interrupt, shadow, auto_psv)) _SI2C1Interrupt(void)
{
	static volatile SAmpCtrl s_Send;
	static volatile UINT8 s_I2CCount = 0;
	UINT8 temp;

	switch(I2C1STAT & 0b000100110)
	{
	case 0b000000110: // 送信(1Byte目)
		// データ(D/A=0) マスターへ送信(R/W=1) 受信完了(RBF=1)
		temp = I2C1RCV;

		//送信データをキャプチャ
		s_Send.Status = s_Status;
		s_Send.Setting = s_Setting;

		I2C1TRN = s_Send.Bytes[0];
		s_I2CCount = 1;
		
		I2C1CONbits.SCLREL = 1;
		break;

	case 0b000100100: // 送信(2Byte目以降)
		// データ(D/A=1) マスターへ送信(R/W=1) 受信完了(RBF=0)
		if (I2C1STATbits.ACKSTAT == 0)
		{
			if (s_I2CCount < sizeof(s_Send.Bytes))
			{
				I2C1TRN = s_Send.Bytes[s_I2CCount];
				++s_I2CCount;
			}
			else
			{
				I2C1TRN = 0;
			}
		}
		I2C1CONbits.SCLREL = 1;
		break;

	case 0b000000010: // 受信(アドレス)
		// データ(D/A=0) マスターから受信(R/W=0) 受信完了(RBF=1)
		temp = I2C1RCV; // 空読み
		s_I2CCount = 0;
		
		I2C1CONbits.SCLREL = 1;
		break;

	case 0b000100010: // 受信(1Byte目以降)
		// データ(D/A=1) マスターから受信(R/W=0) 受信完了(RBF=1)
		temp = I2C1RCV;
		// コマンド受信
		if (!s_Status.bCmdBusy)
		{
			s_Recv[s_I2CCount++] = temp;
			if (s_I2CCount >= 2)
			{
				s_I2CCount = 0;
				s_Status.bCmdBusy = TRUE;
			}
		}
		I2C1CONbits.SCLREL = 1;
		break;

	default:
		temp = I2C1RCV;
		I2C1CONbits.SCLREL = 1;
		break;
	}

	_SI2C1IF = 0;
}

//---------------------------------------------------------------

// ●LMK04832レジスタ設定(22.5792MHz用)
// ※8.5.1 Recommended Programming Sequence
const UINT32 s_LMK04832Regs1[] =
{
	0x000090, // R0 (INIT)
	0x000000, // R0
	0x000200, // R2
	//0x000306, // R3
	//0x0004D1, // R4
	//0x000563, // R5
	//0x000650, // R6
	//0x000C51, // R12
	//0x000D04, // R13
	0x010002, // R256
	0x01010A, // R257
	0x010290, // R258
	0x010350, // R259
	0x010410, // R260
	0x010500, // R261
	0x010601, // R262
	0x010700, // R263
	0x010804, // R264
	0x01090A, // R265
	0x010A90, // R266
	0x010B50, // R267
	0x010C10, // R268
	0x010D00, // R269
	0x010E01, // R270
	0x010F00, // R271
	0x011008, // R272
	0x01110A, // R273
	0x011290, // R274
	0x011350, // R275
	0x011410, // R276
	0x011500, // R277
	0x011601, // R278
	0x011700, // R279
	0x011808, // R280
	0x01190A, // R281
	0x011A90, // R282
	0x011B50, // R283
	0x011C10, // R284
	0x011D00, // R285
	0x011E01, // R286
	0x011F00, // R287
	0x01206E, // R288
	0x01210A, // R289
	0x012270, // R290
	0x012340, // R291
	0x012410, // R292
	0x012500, // R293
	0x012601, // R294
	0x01270A, // R295
	0x012808, // R296
	0x01290A, // R297
	0x012A90, // R298
	0x012B50, // R299
	0x012C10, // R300
	0x012D00, // R301
	0x012E01, // R302
	0x012F00, // R303
	0x013002, // R304
	0x01310A, // R305
	0x013290, // R306
	0x013350, // R307
	0x013410, // R308
	0x013500, // R309
	0x013601, // R310
	0x013700, // R311
	0x013800, // R312
	0x013900, // R313
	0x013A00, // R314
	0x013B0C, // R315
	0x013C00, // R316
	0x013D08, // R317
	0x013E03, // R318
	0x013F0B, // R319
	0x01400F, // R320
	0x014100, // R321
	0x014200, // R322
	0x014300, // R323
	0x014410, // R324
	0x014500, // R325
	0x014609, // R326
	0x01470E, // R327
	0x01480B, // R328
	0x014943, // R329
	0x014A03, // R330
	0x014B32, // R331
	0x014C00, // R332
	0x014D00, // R333
	0x014EC0, // R334
	0x014F7F, // R335
	0x015011, // R336
	0x015102, // R337
	0x015200, // R338
	0x015300, // R339
	0x015414, // R340
	0x015500, // R341
	0x015696, // R342
	0x015700, // R343
	0x015896, // R344
	0x015900, // R345
	0x015A14, // R346
	0x015B1B, // R347
	0x015C20, // R348
	0x015D00, // R349
	0x015E1E, // R350
	0x015F1B, // R351
	0x016000, // R352
	0x016102, // R353
	0x016245, // R354
	0x016300, // R355
	0x016400, // R356
	0x01650B, // R357
	0x017310, // R371 ※この位置に移動する
	0x016959, // R361
	0x016A3F, // R362
	0x016BFF, // R363
	0x016C00, // R364
	0x016D00, // R365
	0x016E03, // R366
	//0x017310, // R371
	0x017700, // R375
	0x018200, // R386
	0x018300, // R387
	0x016600, // R358
	0x016700, // R359
	0x01680B, // R360
	0x055500, // R1365
	0xFFFFFF, // LastMaker
};

// ●LMK04832レジスタ設定(24.57602MHz用)
// ※8.5.1 Recommended Programming Sequence
const UINT32 s_LMK04832Regs2[] =
{
	0x000090, // R0 (INIT)
	0x000000, // R0
	0x000200, // R2
	//0x000306, // R3
	//0x0004D1, // R4
	//0x000563, // R5
	//0x000650, // R6
	//0x000C51, // R12
	//0x000D04, // R13
	0x010002, // R256
	0x01010A, // R257
	0x010290, // R258
	0x010350, // R259
	0x010410, // R260
	0x010500, // R261
	0x010601, // R262
	0x010700, // R263
	0x010804, // R264
	0x01090A, // R265
	0x010A90, // R266
	0x010B50, // R267
	0x010C10, // R268
	0x010D00, // R269
	0x010E01, // R270
	0x010F00, // R271
	0x011008, // R272
	0x01110A, // R273
	0x011290, // R274
	0x011350, // R275
	0x011410, // R276
	0x011500, // R277
	0x011601, // R278
	0x011700, // R279
	0x011808, // R280
	0x01190A, // R281
	0x011A90, // R282
	0x011B50, // R283
	0x011C10, // R284
	0x011D00, // R285
	0x011E01, // R286
	0x011F00, // R287
	0x012064, // R288
	0x01210A, // R289
	0x012270, // R290
	0x012340, // R291
	0x012410, // R292
	0x012500, // R293
	0x012601, // R294
	0x01270A, // R295
	0x012808, // R296
	0x01290A, // R297
	0x012A90, // R298
	0x012B50, // R299
	0x012C10, // R300
	0x012D00, // R301
	0x012E01, // R302
	0x012F00, // R303
	0x013002, // R304
	0x01310A, // R305
	0x013290, // R306
	0x013350, // R307
	0x013410, // R308
	0x013500, // R309
	0x013601, // R310
	0x013700, // R311
	0x013800, // R312
	0x013900, // R313
	0x013A00, // R314
	0x013B0C, // R315
	0x013C00, // R316
	0x013D08, // R317
	0x013E03, // R318
	0x013F0B, // R319
	0x01400F, // R320
	0x014100, // R321
	0x014200, // R322
	0x014300, // R323
	0x014410, // R324
	0x014500, // R325
	0x014609, // R326
	0x01470E, // R327
	0x01480B, // R328
	0x014943, // R329
	0x014A03, // R330
	0x014B32, // R331
	0x014C00, // R332
	0x014D00, // R333
	0x014EC0, // R334
	0x014F7F, // R335
	0x015011, // R336
	0x015102, // R337
	0x015200, // R338
	0x015300, // R339
	0x015414, // R340
	0x015500, // R341
	0x015696, // R342
	0x015700, // R343
	0x015896, // R344
	0x015900, // R345
	0x015A14, // R346
	0x015B1A, // R347
	0x015C20, // R348
	0x015D00, // R349
	0x015E1E, // R350
	0x015F1B, // R351
	0x016000, // R352
	0x016102, // R353
	0x016245, // R354
	0x016300, // R355
	0x016400, // R356
	0x01650A, // R357
	0x017310, // R371 ※この位置に移動する
	0x016959, // R361
	0x016A3F, // R362
	0x016BFF, // R363
	0x016C00, // R364
	0x016D00, // R365
	0x016E03, // R366
	//0x017310, // R371
	0x017700, // R375
	0x018200, // R386
	0x018300, // R387
	0x016600, // R358
	0x016700, // R359
	0x01680A, // R360
	0x055500, // R1365
	0xFFFFFF, // LastMaker
};
