/****************************************************************************

 ■HiResAmp 表示制御(DisplayControl)
		Copyright(C) 2020 電子うさぎ https://電子うさぎ.com/

****************************************************************************/

// PIC32MZ2048EFH064 Configuration Bit Settings

// DEVCFG3
#pragma config USERID = 0xFFFF          // Enter Hexadecimal value (Enter Hexadecimal value)
#pragma config FMIIEN = ON              // Ethernet RMII/MII Enable (MII Enabled)
#pragma config FETHIO = ON              // Ethernet I/O Pin Select (Default Ethernet I/O)
#pragma config PGL1WAY = ON             // Permission Group Lock One Way Configuration (Allow only one reconfiguration)
#pragma config PMDL1WAY = ON            // Peripheral Module Disable Configuration (Allow only one reconfiguration)
#pragma config IOL1WAY = ON             // Peripheral Pin Select Configuration (Allow only one reconfiguration)
#pragma config FUSBIDIO = OFF           // USB USBID Selection (Controlled by Port Function)

// DEVCFG2
#pragma config FPLLIDIV = DIV_1         // System PLL Input Divider (1x Divider)
#pragma config FPLLRNG = RANGE_5_10_MHZ // System PLL Input Range (5-10 MHz Input)
#pragma config FPLLICLK = PLL_FRC       // System PLL Input Clock Selection (FRC is input to the System PLL)
#pragma config FPLLMULT = MUL_50        // System PLL Multiplier (PLL Multiply by 50)
#pragma config FPLLODIV = DIV_2         // System PLL Output Clock Divider (2x Divider)
#pragma config UPLLFSEL = FREQ_24MHZ    // USB PLL Input Frequency Selection (USB PLL input is 24 MHz)

// DEVCFG1
#pragma config FNOSC = SPLL             // Oscillator Selection Bits (System PLL)
#pragma config DMTINTV = WIN_127_128    // DMT Count Window Interval (Window/Interval value is 127/128 counter value)
#pragma config FSOSCEN = OFF            // Secondary Oscillator Enable (Disable SOSC)
#pragma config IESO = ON                // Internal/External Switch Over (Enabled)
#pragma config POSCMOD = OFF            // Primary Oscillator Configuration (Primary osc disabled)
#pragma config OSCIOFNC = OFF           // CLKO Output Signal Active on the OSCO Pin (Disabled)
#pragma config FCKSM = CSECME           // Clock Switching and Monitor Selection (Clock Switch Enabled, FSCM Enabled)
#pragma config WDTPS = PS8192           // Watchdog Timer Postscaler (1:8192)
#pragma config WDTSPGM = STOP           // Watchdog Timer Stop During Flash Programming (WDT stops during Flash programming)
#pragma config WINDIS = NORMAL          // Watchdog Timer Window Mode (Watchdog Timer is in non-Window mode)
#pragma config FWDTEN = OFF             // Watchdog Timer Enable (WDT Disabled)
#pragma config FWDTWINSZ = WINSZ_25     // Watchdog Timer Window Size (Window size is 25%)
#pragma config DMTCNT = DMT31           // Deadman Timer Count Selection (2^31 (2147483648))
#pragma config FDMTEN = OFF             // Deadman Timer Enable (Deadman Timer is disabled)

// DEVCFG0
#pragma config DEBUG = OFF              // Background Debugger Enable (Debugger is disabled)
#pragma config JTAGEN = OFF             // JTAG Enable (JTAG Disabled)
#pragma config ICESEL = ICS_PGx1        // ICE/ICD Comm Channel Select (Communicate on PGEC1/PGED1)
#pragma config TRCEN = ON               // Trace Enable (Trace features in the CPU are enabled)
#pragma config BOOTISA = MIPS32         // Boot ISA Selection (Boot code and Exception code is MIPS32)
#pragma config FECCCON = OFF_UNLOCKED   // Dynamic Flash ECC Configuration (ECC and Dynamic ECC are disabled (ECCCON bits are writable))
#pragma config FSLEEP = OFF             // Flash Sleep Mode (Flash is powered down when the device is in Sleep mode)
#pragma config DBGPER = PG_ALL          // Debug Mode CPU Access Permission (Allow CPU access to all permission regions)
#pragma config SMCLR = MCLR_NORM        // Soft Master Clear Enable bit (MCLR pin generates a normal system Reset)
#pragma config SOSCGAIN = GAIN_2X       // Secondary Oscillator Gain Control bits (2x gain setting)
#pragma config SOSCBOOST = ON           // Secondary Oscillator Boost Kick Start Enable bit (Boost the kick start of the oscillator)
#pragma config POSCGAIN = GAIN_2X       // Primary Oscillator Gain Control bits (2x gain setting)
#pragma config POSCBOOST = ON           // Primary Oscillator Boost Kick Start Enable bit (Boost the kick start of the oscillator)
#pragma config EJTAGBEN = NORMAL        // EJTAG Boot (Normal EJTAG functionality)

// DEVCP0
#pragma config CP = OFF                 // Code Protect (Protection Disabled)

// SEQ3
#pragma config TSEQ = 0xFFFF            // Boot Flash True Sequence Number (Enter Hexadecimal value)
#pragma config CSEQ = 0xFFFF            // Boot Flash Complement Sequence Number (Enter Hexadecimal value)

#include <xc.h>
#include <proc/p32mz2048efh064.h>
#include <GenericTypeDefs.h>
#include <math.h>

//=================================================================

#define DEBUG_LED		LATCbits.LATC15
#define COM_SCL			PORTDbits.RD10
#define COM_SDA			PORTDbits.RD9
#define COM_SCL_TRIS	TRISDbits.TRISD10

#define LCD_SetRST()	LATDSET = _LATD_LATD5_MASK
#define LCD_SetCS()		LATDSET = _LATD_LATD1_MASK
#define LCD_SetRS()		LATDSET = _LATD_LATD2_MASK
#define LCD_SetWR()		LATDSET = _LATD_LATD3_MASK
#define LCD_SetRD()		LATDSET = _LATD_LATD4_MASK
#define LCD_SetLED()	LATCSET = _LATC_LATC14_MASK
#define LCD_ClrRST()	LATDCLR = _LATD_LATD5_MASK
#define LCD_ClrCS()		LATDCLR = _LATD_LATD1_MASK
#define LCD_ClrRS()		LATDCLR = _LATD_LATD2_MASK
#define LCD_ClrWR()		LATDCLR = _LATD_LATD3_MASK
#define LCD_ClrRD()		LATDCLR = _LATD_LATD4_MASK
#define LCD_ClrLED()	LATCCLR = _LATC_LATC14_MASK

#define LCD_DirInput()	TRISE = 0xFFFFFFFF
#define LCD_DirOutput()	TRISE = 0xFFFFFF00
#define LCD_DATA_I		PORTE
#define LCD_DATA_O		LATE

#define LCD_WIDTH		400
#define LCD_HEIGHT		240

// 表示モード
#define DISP_INIT		0	// 初期画面
#define DISP_HOME		1	// ホーム画面
#define DISP_OUTPUT		2	// 出力切替え
#define DISP_BBOOST		3	// バスブースト設定
#define DISP_DACGAIN	4	// DACゲイン設定
#define DISP_PCMFLT1	5	// PCMフィルタ設定
#define DISP_PCMFLT2	6	// PCMフィルタ設定
#define DISP_PCMFLT3	7	// PCMフィルタ設定
#define DISP_PCMFLT4	8	// PCMフィルタ設定
#define DISP_PCMFLT5	9	// PCMフィルタ設定
#define DISP_DSDFLT		10	// DSDフィルタ設定
#define DISP_STATUS1	11	// ステータス表示
#define DISP_STATUS2	12	// ステータス表示
#define DISP_STATUS3	13	// ステータス表示
#define DISP_MUTE		14	// ミュート
#define DISP_SHUTDOWN	99	// シャットダウン

// 数値描画位置
#define DN_LEFT			0	// 左寄せ
#define DN_CENTER		1	// 中央
#define DN_RIGHT		2	// 右寄せ

// ステート値(アンプ制御)
#define STATE_INIT			0		// 初期化中(共通)
#define STATE_STANDBY		1		// 待機中
#define STATE_CONNECTED		2		// 接続済
#define STATE_PLAYING		3		// 再生中
#define STATE_ERROR			4		// エラー発生
#define STATE_SHUTDOWN		5		// シャットダウン(共通)

// 再生フォーマット
#define AF_UNKNOWN			0
#define AF_I2S_1624			1
#define AF_I2S_32			2
#define AF_DSD_64			3
#define AF_DSD_128			4
#define AF_DSD_256			5
#define AF_DSD_512			6

// 入力制御定義
#define COMBO384_FREQ_MASK	0x0F
#define COMBO384_DSDON_MASK	0x40
#define COMBO384_DSD64_MASK	0x20
#define COMBO384_MUTE_MASK	0x10

#define DSD_UNKNOWN			0
#define DSD_32KHZ			1		
#define DSD_44KHZ			2
#define DSD_48KHZ			3

// 外部PC管理状態
#define PC_NONE				0	// PC管理なし
#define PC_ENABLE			1	// 制御アプリ認識完了
#define PC_ERROR			2	// 制御アプリエラー

// ボリュームポイント描画関連
#define MUTE_DEG			226.1233027			// ミュートポイントの角度
#define VMAX_DEG			(180.0 - MUTE_DEG)	// ボリューム最大位置の角度
#define VRNG_DEG			(MUTE_DEG - VMAX_DEG)
#define VOL_MAX				72
#define PI					3.141592654
#define VC_R				74.0	// ボリューム円の半径
#define VC_X				321		// ボリューム円中心のスクリーン座標(X)
#define VC_Y				118		// ボリューム円中心のスクリーン座標(Y)
#define VPOS_W				10		// ボリュームポイント画像幅
#define VPOS_H				10		// ボリュームポイント画像高

//---------------------------------------------------------------

// 画像データ
typedef struct {
	int w;
	int h;
	const BYTE* p;
} SBitmap;

// ボリューム位置画像関連
typedef struct {
	BYTE Mixed[VPOS_W * VPOS_H * 3];
	BYTE Saved[VPOS_W * VPOS_H * 3];
	int nPosX;
	int nPosY;
} SVolPos;

#pragma pack(push)
#pragma pack(1)
// アンプ制御と共通↓↓↓↓↓
// LMK04832ステータス
typedef struct {
	UINT8 LOS			: 1;
	UINT8 PLL1_LD		: 1;
	UINT8 PLL2_LD		: 1;
	UINT8 PLL1_LD_LOST	: 1;
	UINT8 PLL2_LD_LOST	: 1;
	UINT8 HOLDOVER		: 1;
	UINT8 DAC_LOCKED	: 1;
	UINT16 DAC_VALUE;
} SLMK04832;

// バイアス制御ステータス
typedef struct {
	UINT8 nStatus;
	UINT8 nDacVal;
	UINT16 nAvgCur;
	UINT16 nSetCur;
} SBiasCtrl;

// ステータス総合
typedef struct {
	UINT8 bCmdBusy	: 1;	// コマンドBusy
	UINT8 nState	: 3;	// アンプ制御の状態
	UINT8 nErrInfo	: 4;	// エラー状態時の付加情報
	UINT8 nFormat	: 7;	// 再生フォーマット
	UINT8 nVCXO		: 1;	// VCXO周波数セレクト
	UINT8 nInputs;			// COMBO384 IO値
	UINT8 nInputInfo;		// 量子化ビット数又はDDSサンプリング周波数
	UINT8 nAnaATT;			// アナログアッテネータ
	UINT8 nDigATT;			// デジタルアッテネータ
	UINT8 bAdvValid;		// 検出電圧値有効
	UINT16 nAdvValue;		// 検出電圧値
	SLMK04832 LMK04832;
	SBiasCtrl BiasCtrl[4];
} SStatus;

// 動作設定
typedef struct {
	UINT8 bEnable;			// 0or1
	UINT16 nVolume;			// 音量設定時Volume
	UINT16 nAdcVal;			// 音量設定値(DAC値)	
} SAutoVolCtrl;

typedef struct {
	UINT8 nVolume[2];		// HP=0～72, SP=0～81
	UINT8 nBBoost[2];		// 0or1
	UINT8 nOutput	: 1;	// 0:HP, 1:SP
	UINT8 nMute		: 1;	// 0or1
	UINT8 nPcmFlt	: 3;	// 0～6(4と5は同じ)
	UINT8 nDsdFlt	: 1;	// 0or1
	UINT8 nDacGain[2];		// 0or1
	SAutoVolCtrl Avc[2];
} SSetting;
// アンプ制御と共通↑↑↑↑↑

// 全体制御と共通↓↓↓↓↓
// 表示設定
typedef struct {
	UINT8 nMode;		// 画面タイプ
	UINT8 bDone	:	1;	// Doneフラグ
	UINT8 nPCState:	4;	// 外部PC管理状態
} SDisplay;

// 送信データ
typedef union {
	struct {
		SStatus Status;
		SSetting Setting;
		SDisplay Display;
		UINT8 nEndMarker; 
	};
	UINT8 Bytes[sizeof(SStatus) + sizeof(SSetting) + sizeof(SDisplay) + 1];
} SDispCtrl;
// 全体制御と共通↑↑↑↑↑
#pragma pack(pop)

//---------------------------------------------------------------

static void SendI2C(UINT8 nAddr, const UINT8* pData, int nCount);
static void RecvI2C(UINT8 nAddr, UINT8* pData, int nCount);

static void InitSystem(void);
static void PerformanceConfig(unsigned int sysclk);
static void InitApp(void);
static void CreateVolPoints(void);
static void Run(void);
static void DrawHome(void);
static void DrawBias(BOOL bAll);
static void DrawMenu(int nMode);
static void DrawStatus(int nMode);
static void Shutdown(void);
static int DrawNumber(int nFont, int x, int y, int w, int nVal, int nAdj);

static void ResetLCD(void);
static void InitLCD(void);
static void FillRect(int x, int y, int w, int h, UINT32 color);
static void DrawBitmap(int x, int y, SBitmap* pbm);
static UINT16 ReadDeviceID(void);
static void WriteCtrlReg(UINT16 addr, UINT16 data);
static UINT16 ReadReg(BOOL index);
static void WriteReg(BOOL index, UINT16 data);

#define RGB(r, g, b)	(((r) << 16) | ((g) << 8) | ((b) << 0))
#define GET_R(x)		((BYTE)((x) >> 16))
#define GET_G(x)		((BYTE)((x) >> 8))
#define GET_B(x)		((BYTE)((x) >> 0))
#define ClearRect(x, y, w, h)	FillRect(x, y, w, h, RGB(0, 0, 0))

static void Delay(register unsigned long time);

#define DELAY_UNIT		10	// 10ns
#define DelayNs(x)		Delay((x) / DELAY_UNIT)
#define DelayUs(x)		Delay((x) * (1000 / DELAY_UNIT))
#define DelayMs(x)		Delay((x) * (1000000 / DELAY_UNIT))

#define IsBusy()		(!COM_SCL_TRIS)
#define ClearBusy()		COM_SCL_TRIS = 1

#define EnableWdt()		WDTCONbits.ON = 1
#define ClrWdt()		WDTCONbits.WDTCLRKEY = 0x5743

//---------------------------------------------------------------

// 画像データ
#include "bitmaps1.h"
#include "bitmaps2.h"
#include "bitmaps3.h"
#include "bitmaps4.h"

static volatile SDispCtrl s_CtrlData = { 0 };
static SDispCtrl s_Current;
static SVolPos s_VolPos[VOL_MAX];
static volatile BOOL s_bSens = FALSE;

//=================================================================

// ●メインエントリ
void main(void)
{
	InitSystem();
	InitApp();
	Run();
	Shutdown();
}

// ●システム初期化
static void InitSystem(void)
{
 	// SYSCLK=200MHz駆動
	PerformanceConfig(200000000UL);
	
	// ポート初期化
	PORTB = PORTC = PORTD = PORTE = PORTF = PORTG = 0;
	LATB = LATC = LATD = LATE = LATF = LATG = 0;
	
	// DEBUG_LED(RC15)
	TRISCbits.TRISC15 = 0;
	DEBUG_LED = 1;
	
	// LCD RST(RD5)
	LCD_SetRST();
	TRISDbits.TRISD5 = 0;
	
	// LCD CS(RD1)
	LCD_SetCS();
	TRISDbits.TRISD1 = 0;

	// LCD RS(RD2)
	TRISDbits.TRISD2 = 0;

	// LCD WR(RD3)
	LCD_SetWR();
	TRISDbits.TRISD3 = 0;

	// LCD RD(RD4)
	LCD_SetRD();
	TRISDbits.TRISD4 = 0;
	
	// LCD_LED(RB2)
	TRISCbits.TRISC14 = 0;
	
	// LCD_DATA
	TRISE = 0xFFFFFFFF;
	ANSELE = 0xFFFFFF00;
	
	// Timer2/3
	T2CONbits.T32 = 1;
	PR2 = 75000000; // 800ms
	IPC3bits.T3IP = 4;
	IEC0bits.T3IE = 1;
	
	// オリジナルプロトコル通信(SCL:RD10, SDA:RD9)
	COM_SCL_TRIS = 0; // busy
	CNCONDbits.EDGEDETECT = 1;
	CNNEDbits.CNNED10 = 1;
	IPC30bits.CNDIP = 6;
	IEC3bits.CNDIE = 1;
	CNCONDbits.ON = 1;
	
	// 割り込み許可
 	INTCONbits.MVEC = 1; // マルチベクタ割り込み
   __builtin_enable_interrupts();	
}

// ●PIC32MZ EF パフォーマンス設定 (SYSTEMConfigPerformance相当)
static void PerformanceConfig(unsigned int sysclk)
{		
	// TABLE A-4: CPU DIFFERENCES
   
	// PBDIVの設定
	// →PIC32MZ EFではPBxDIVに変更された SYSCLK=200MHzの場合考慮不要
	
	// RAMアクセスWait(BMXWSDRMで設定)
	// →PIC32MZ EFでは不要になった

	// Predictive Prefetch Enable bits
	PRECONbits.PREFEN = 0b11; // Enable predictive prefetch for any address

    // PFM Access Time Defined in Terms of SYSCLK Wait States bits
	// TABLE 37-13: DC CHARACTERISTICS: PROGRAM FLASH MEMORY WAIT STATES (With ECC)
	if (sysclk <= 60000000)
		PRECONbits.PFMWS = 0;
	else if ((60000000 < sysclk) && (sysclk <= 120000000))
		PRECONbits.PFMWS = 1;
	else
		PRECONbits.PFMWS = 2;
		
    // CP0 REGISTER 16, SELECT 0 - Kseg0 Coherency Algorithm bits 
	unsigned long r16_0 = __builtin_mfc0(16, 0);
    r16_0 = (r16_0 & ~0b111) | 0b011; // Cacheable, non-coherent, write-back, write allocate
	__builtin_mtc0(16, 0, r16_0);
}

// ●アプリ初期化
static void InitApp(void)
{
#ifndef _DEBUG
	EnableWdt();
#endif

	int i = 0;
	do s_CtrlData.Bytes[i] = 0;
	while(++i < sizeof(s_CtrlData.Bytes));
	s_Current = s_CtrlData;
	
	CreateVolPoints();
	
	// 安定待ち
	DelayMs(100);

	// LCD(M032C9326D8)のリセット
	ResetLCD();
	
	// デバイスIDをチェックして動作確認
	if (ReadDeviceID() != 0x9326)
		DEBUG_LED = 0;

	// LCDの初期化
	InitLCD();

	// 初期表示
	DrawBitmap(-1, -1, &bm_initial);

	// 通信受付開始
	ClearBusy();
}

// ●ボリュームポイントのイメージを作成する
static void CreateVolPoints(void)
{
	int nPos;
	for(nPos = 0; nPos < VOL_MAX; nPos++)
	{
		SVolPos* pVolPos = &s_VolPos[nPos];
		
		// 描画位置を算出する
		double dVol = MUTE_DEG - (VRNG_DEG * (nPos + 1)) / VOL_MAX;
		dVol = dVol * PI / 180.0; // deg → rad
		int rx = cos(dVol) * VC_R;
		int ry = sin(dVol) * VC_R;
	
		int x = VC_X + rx - VPOS_W / 2;
		int y = VC_Y - ry - VPOS_H / 2;
		pVolPos->nPosX = x;
		pVolPos->nPosY = y;

		// 背景と簡易クロマキー
		BYTE* pMix = pVolPos->Mixed;
		BYTE* pSbk = pVolPos->Saved;
		const BYTE* pMark = bm_volpos.p;

		int i, j;
		for(i = 0; i < VPOS_H; i++)
		{
			const BYTE* pHbk = bm_homebk.p + ((y + i) * bm_homebk.w + x) * 3;
			
			for(j = 0; j < (VPOS_W * 3); j++, pMix++, pSbk++, pMark++, pHbk++)
			{
				*pMix = (*pMark == 0)? *pHbk : *pMark; // 色値0は背景値を使用
				*pSbk = *pHbk;
			}
		}
	}
}

// ●実行ループ
static void Run(void)
{
	// シャットダウンまで無限ループ
	for(;;)
	{
		ClrWdt();

		if (IsBusy())
		{
			switch (s_CtrlData.Display.nMode)
			{
			case DISP_HOME:
				DrawHome();
				break;

			case DISP_SHUTDOWN:
				// PowerOFF操作 →Busyのまま抜ける
				return;
			
			default:
				DrawMenu(s_CtrlData.Display.nMode);
				break;
			}
			
			s_Current = s_CtrlData;
			
			ClearBusy();
		}
		else
		{
			// 何もすることがない間はアイドルモード
			_wait();
		}
	}
}

// ●ホーム画面
static void DrawHome(void)
{
	static SBitmap* pTitles[] = {
		NULL, &bm_title_nousb, &bm_title_usbok, &bm_title_playing, &bm_title_error, NULL };

	static SBitmap* pFreqs[] = { &bm_pcmfreq_32, &bm_pcmfreq_44, &bm_pcmfreq_48, &bm_pcmfreq_88,
		&bm_pcmfreq_96, &bm_pcmfreq_176, &bm_pcmfreq_192, &bm_pcmfreq_352, &bm_pcmfreq_384,
		NULL, NULL, NULL, NULL, NULL, NULL, NULL};

	SBitmap* pBm;
	int i, x, y, n;

	// 背景
	BOOL bAll = FALSE;
	if (s_Current.Display.nMode != DISP_HOME)
	{
		DrawBitmap(0, 0, &bm_homebk);
		bAll = TRUE;
	}
	
	if (bAll || s_Current.Status.nState != s_CtrlData.Status.nState)
	{
		// タイトルアイコン
		if (s_CtrlData.Status.nState == STATE_PLAYING)
			pBm = &bm_topicon_red;
		else
			pBm = &bm_topicon_gry;

		DrawBitmap(3, 14, pBm);

		// タイトル文字列
		pBm = pTitles[s_CtrlData.Status.nState];
		if (pBm != NULL)
			DrawBitmap(30, 10, pBm);
		else
			ClearRect(30, 10, bm_title_playing.w, bm_title_playing.h);
	}
	
	// ロックアイコン
	if (bAll || (s_Current.Status.LMK04832.PLL1_LD != s_CtrlData.Status.LMK04832.PLL1_LD)
		|| (s_Current.Status.LMK04832.PLL2_LD != s_CtrlData.Status.LMK04832.PLL2_LD))
	{
		if (s_CtrlData.Status.LMK04832.PLL1_LD && s_CtrlData.Status.LMK04832.PLL2_LD)
			DrawBitmap(232, 18, &bm_lock);
		else
			ClearRect(232, 18, bm_lock.w, bm_lock.h);
	}
	
	// PCアイコン
	if (bAll || (s_Current.Display.nPCState != s_CtrlData.Display.nPCState))
	{
		if (s_CtrlData.Display.nPCState == PC_ENABLE)
			DrawBitmap(320, 18, &bm_pc_nrm);
		else if (s_CtrlData.Display.nPCState == PC_ERROR)
			DrawBitmap(320, 18, &bm_pc_err);
		else
			ClearRect(320, 18, bm_pc_nrm.w, bm_pc_nrm.h);
	}
	
	// オーディオフォーマット
	if (bAll || (s_Current.Status.nFormat != s_CtrlData.Status.nFormat)
			|| (s_Current.Status.nInputInfo != s_CtrlData.Status.nInputInfo))
	{
		x = 22; y = 66;
		switch(s_CtrlData.Status.nFormat)
		{
		case AF_I2S_1624: case AF_I2S_32:
			// PCM
			DrawBitmap(x, y, &bm_atype_pcm);
			x += bm_atype_pcm.w;
			ClearRect(x, y, 8, 20);
			x += 8;
			// 周波数
			i = s_CtrlData.Status.nInputs & COMBO384_FREQ_MASK;
			pBm = pFreqs[i];
			if (pBm != NULL)
			{
				DrawBitmap(x, y, pBm);
				x += pBm->w;
				//kHz
				DrawBitmap(x, y, &bm_pcmfreq_khz);
				x += bm_pcmfreq_khz.w;
				ClearRect(x, y, 8, 20);
				x += 8;
			}
			// ビット数
			switch(s_CtrlData.Status.nInputInfo)
			{
			case 16: pBm = &bm_pcmbit_16bit; break;
			case 24: pBm = &bm_pcmbit_24bit; break;
			case 32: pBm = &bm_pcmbit_32bit; break;
			default: pBm = NULL; break;
			}
			if (pBm != NULL)
			{
				DrawBitmap(x, y, pBm);
				x += pBm->w;
			}
			break;

		case AF_DSD_64:
			DrawBitmap(x, y, &bm_dsd64);
			x += bm_dsd64.w;
			ClearRect(x, y, 8, 20);
			x += 8;
			break;
		case AF_DSD_128:
			DrawBitmap(x, y, &bm_dsd128);
			x += bm_dsd128.w;
			ClearRect(x, y, 8, 20);
			x += 8;
			break;
		case AF_DSD_256:
			DrawBitmap(x, y, &bm_dsd256);
			x += bm_dsd256.w;
			ClearRect(x, y, 8, 20);
			x += 8;
			break;
		case AF_DSD_512:
			DrawBitmap(x, y, &bm_dsd512);
			x += bm_dsd512.w;
			ClearRect(x, y, 8, 20);
			x += 8;
			break;
		}
		if (s_CtrlData.Status.nFormat >= AF_DSD_64)
		{
			switch(s_CtrlData.Status.nInputInfo)
			{
			case DSD_32KHZ: pBm = &bm_pcmfreq_32; break;
			case DSD_44KHZ: pBm = &bm_pcmfreq_44; break;
			case DSD_48KHZ: pBm = &bm_pcmfreq_48; break;
			default: pBm = NULL; break;
			}
			if (pBm != NULL)
			{
				DrawBitmap(x, y, pBm);
				x += pBm->w;
				//kHz
				DrawBitmap(x, y, &bm_pcmfreq_khz);
				x += bm_pcmfreq_khz.w;
			}
		}
		ClearRect(x, y, 242 - x, 20);
	}
	
	// ボリューム情報(ATT)
	if (bAll || (s_Current.Status.nAnaATT != s_CtrlData.Status.nAnaATT))
	{
		x = 173, y = 104;
		n = s_CtrlData.Status.nAnaATT * 12;
		x += DrawNumber(0, x, y, 0, n, DN_LEFT);
		DrawBitmap(x + 1, y, &bm_w14_db);
		x += 1 + bm_w14_db.w;
		ClearRect(x, y, 220 - x, bm_w14_db.h);
	}
	if (bAll || (s_Current.Status.nDigATT != s_CtrlData.Status.nDigATT))
	{
		x = 173, y = 126;
		if (s_CtrlData.Status.nDigATT > 0)
		{
			n = 1275 - s_CtrlData.Status.nDigATT * 5;
			x += DrawNumber(0, x, y, 0, n / 10, DN_LEFT);
			n = n % 10;
			if (n != 0)
			{
				DrawBitmap(x, y, &bm_dig_w14_dot);
				x += bm_dig_w14_dot.w;
				x += DrawNumber(0, x, y, 0, n, DN_LEFT);
			}
			DrawBitmap(x + 1, y, &bm_w14_db);
			x += 1 + bm_w14_db.w;
		}
		else
		{
			DrawBitmap(x, y, &bm_inbig);
			x += bm_inbig.w;
		}
		ClearRect(x, y, 235 - x, bm_w14_db.h);
	}

	// バイアス情報
	DrawBias(bAll);

	// オートボリューム(AD値)
	if (bAll || (s_Current.Status.nAdvValue != s_CtrlData.Status.nAdvValue)
			|| (s_Current.Status.bAdvValid != s_CtrlData.Status.bAdvValid))
	{
		x = 250, y = 214;
		if (s_CtrlData.Status.bAdvValid)
		{
			DrawBitmap(x, y, &bm_adinfo);
			x += bm_adinfo.w + 2;
			x += DrawNumber(2, x, y, 0, s_CtrlData.Status.nAdvValue, DN_LEFT);
		}
		ClearRect(x, y, 335 - x, bm_adinfo.h);
	}

	BOOL bOutput = FALSE;
	int nOutput = s_CtrlData.Setting.nOutput;

	// 出力アイコン
	if (bAll || (s_Current.Setting.nOutput != s_CtrlData.Setting.nOutput))
	{
		pBm = (s_CtrlData.Setting.nOutput == 0)? &bm_headphone : &bm_speaker;
		DrawBitmap(362, 11, pBm);
		bOutput = TRUE;
	}

	// ボリューム情報(DAC Gain)
	if (bOutput || s_Current.Setting.nDacGain[nOutput] != s_CtrlData.Setting.nDacGain[nOutput])
	{
		pBm = (s_CtrlData.Setting.nDacGain[nOutput] == 0)? &bm_w14_hight : &bm_w14_low;
		x = 167, y = 149;
		DrawBitmap(167, 149, pBm);
		x += pBm->w;
		ClearRect(x, y, 220 - x, pBm->h);
	}
	
	// ボリューム
	if (bOutput ||(s_Current.Setting.nVolume[nOutput] != s_CtrlData.Setting.nVolume[nOutput])
			|| (s_Current.Setting.nMute != s_CtrlData.Setting.nMute))
	{
		x = 270, y = 82;
		if (bOutput || (s_Current.Setting.nVolume[nOutput] >= 10 && s_CtrlData.Setting.nVolume[nOutput] < 10))
			ClearRect(x, y, 100, 65);
			
		DrawNumber(3, x, y, 100, s_CtrlData.Setting.nVolume[nOutput], DN_CENTER);
		
		static int s_nLatest = 0; 
		if (!bAll && (s_nLatest > 0))
		{
			SVolPos* pVolPos = &s_VolPos[s_nLatest - 1];

			SBitmap bm = bm_volpos;
			bm.p = pVolPos->Saved;
			DrawBitmap(pVolPos->nPosX, pVolPos->nPosY, &bm);
		}

		s_nLatest = s_CtrlData.Setting.nVolume[nOutput];
		if (s_nLatest > 0 && !s_CtrlData.Setting.nMute)
		{
			SVolPos* pVolPos = &s_VolPos[s_nLatest - 1];

			SBitmap bm = bm_volpos;
			bm.p = pVolPos->Mixed;
			DrawBitmap(pVolPos->nPosX, pVolPos->nPosY, &bm);
		}
	}
	
	// オートボリューム(設定)
	SAutoVolCtrl* pOldAvc = &s_Current.Setting.Avc[nOutput];
	volatile SAutoVolCtrl* pNewAvc = &s_CtrlData.Setting.Avc[nOutput];
	
	if (bOutput || (pOldAvc->bEnable != pNewAvc->bEnable) || (pOldAvc->nAdcVal != pNewAvc->nAdcVal)
			 || (s_Current.Status.bAdvValid != s_CtrlData.Status.bAdvValid)
				|| (s_Current.Status.nFormat != s_CtrlData.Status.nFormat))
	{
		x = 295; y = 157;
		if (pNewAvc->bEnable)
		{
			if (!s_CtrlData.Status.nFormat == AF_UNKNOWN)
			{
				if (!s_CtrlData.Status.bAdvValid || pNewAvc->nAdcVal == 0)
				{
					DrawBitmap(x, y, &bm_sens);

					// 点滅表示開始
					s_bSens = TRUE;
					T2CONbits.ON = 1;
				}
				else
				{
					T2CONbits.ON = 0;
					DrawBitmap(x, y, &bm_auto);
				}
			}
			else
			{
				T2CONbits.ON = 0;
				DrawBitmap(x, y, &bm_auto_gray);
			}
		}
		else
		{
			T2CONbits.ON = 0;
			ClearRect(x, y, bm_auto.w, bm_auto.h);
		}
	}
	
	// バスブースト
	if (bOutput || (s_Current.Setting.nBBoost[nOutput] != s_CtrlData.Setting.nBBoost[nOutput]))
	{
		if (s_CtrlData.Setting.nBBoost[nOutput])
			DrawBitmap(344, 211, &bm_bboost);
		else
			ClearRect(344, 211, bm_bboost.w, bm_bboost.h);
	}
}

// ●ホーム画面(バイアス情報)
static void DrawBias(BOOL bAll)
{
	volatile SBiasCtrl *pOldP, *pNewP;
	volatile SBiasCtrl *pOldN, *pNewN;
	int x, y, n, j;
	
	// バイアス情報は LP LN RP RN の順
	int i = 0;
	do {
		pOldP = &s_Current.Status.BiasCtrl[i * 2];
		pOldN = pOldP + 1;
		pNewP = &s_CtrlData.Status.BiasCtrl[i * 2];
		pNewN = pNewP + 1;
		
		// カレント電流
		if (bAll || (pOldP->nStatus != pNewP->nStatus) || (pOldN->nStatus != pNewN->nStatus)
				|| (pOldP->nAvgCur != pNewP->nAvgCur) || (pOldN->nAvgCur != pNewN->nAvgCur))
		{
			x = 175;
			y = 197 + (22 * i);

			if (pNewP->nStatus < 7 && pNewN->nStatus < 7) // 7:DAC0なのに下限に達しない
			{
				n = (pNewP->nAvgCur + pNewN->nAvgCur) / 2;
				x += DrawNumber(1, x, y, 0, n, DN_LEFT);
				DrawBitmap(x, y, &bm_w14_ma);
				x += bm_w14_ma.w;
			}
			else
			{
				// どちらかがエラーステータスの場合はError表示
				DrawBitmap(x, y, &bm_error);
				x += bm_error.w;
			}
			ClearRect(x, y, 245 - x, bm_w14_ma.h);
		}

		// DAC値ゲージ(Positive)
		if (bAll || (pOldP->nStatus != pNewP->nStatus) || (pOldP->nDacVal != pNewP->nDacVal))
		{
			x = 34, y = 195 + i * 23;
			FillRect(x, y, 8, 8, (pNewP->nStatus == 2)? RGB(32, 160, 32) : RGB(214, 178, 0));
			x += 9;

			n = (pNewP->nDacVal + 3) / 4;
			for(j = 0; j < 64; j++, x += 2)
				FillRect(x, y, 1, 8, (j >= n)? RGB(112, 112, 112) : RGB(152, 219, 224));
		}

		// DAC値ゲージ(Negative)
		if (bAll || (pOldN->nStatus != pNewN->nStatus) || (pOldN->nDacVal != pNewN->nDacVal))
		{
			x = 34, y = 204 + i * 23;
			FillRect(x, y, 8, 8, (pNewN->nStatus == 2)? RGB(32, 160, 32) : RGB(214, 178, 0));
			x += 9;

			n = (pNewN->nDacVal + 3) / 4;
			for(j = 0; j < 64; j++, x += 2)
				FillRect(x, y, 1, 8, (j >= n)? RGB(112, 112, 112) : RGB(152, 219, 224));
		}
	}
	while(++i < 2);
}

// ●メニュー画面
static void DrawMenu(int nMode)
{
	static int s_nLastNo = -1;
	SBitmap* pBm;
	int i, x, nFlt, nCur;
	
	if (s_Current.Display.nMode != nMode)
	{
		// 画面消去
		if (s_Current.Display.nMode <= DISP_HOME)
		{
			T2CONbits.ON = 0; // 点滅表示停止
	
			ClearRect(0, 0,LCD_WIDTH, LCD_HEIGHT);
			s_nLastNo = -1;
		}
		else
		{
			if ((DISP_PCMFLT1 <= s_Current.Display.nMode) && (s_Current.Display.nMode <= DISP_PCMFLT5))
			{
				if (nMode < DISP_PCMFLT1 || DISP_PCMFLT5 < nMode)
					ClearRect(41, 120, 345, 51);
			}
			else if ((DISP_STATUS1 <= s_Current.Display.nMode) && (s_Current.Display.nMode <= DISP_STATUS3))
				ClearRect(45, 70, 325, 154);
			else
				ClearRect(64, 123, 285, 42);
		}

		SBitmap* pTitle;
		int nNo;
		switch(nMode)
		{
		case DISP_OUTPUT: // 出力切替え
			pTitle = &bm_menu_ou;
			nNo = 1;
			break;
		case DISP_BBOOST: // バスブースト設定
			pTitle = &bm_menu_bb;
			nNo = 2;
			break;
		case DISP_DACGAIN: // DACゲイン設定
			pTitle = &bm_menu_dg;
			nNo = 3;
			break;
		case DISP_PCMFLT1: // PCMフィルタ設定
		case DISP_PCMFLT2: // PCMフィルタ設定
		case DISP_PCMFLT3: // PCMフィルタ設定
		case DISP_PCMFLT4: // PCMフィルタ設定
		case DISP_PCMFLT5: // PCMフィルタ設定
			pTitle = &bm_menu_pf;
			nNo = 4;
			break;
		case DISP_DSDFLT: // DSDフィルタ設定
			pTitle = &bm_menu_df;
			nNo = 5;
			break;
		case DISP_STATUS1: // ステータス表示
			pTitle = &bm_menu_st;
			DrawBitmap(45, 70, &bm_menu_st1bk);
			nNo = 6;
			break;
		case DISP_STATUS2: // ステータス表示
			pTitle = &bm_menu_st;
			DrawBitmap(45, 70, &bm_menu_st2bk);
			nNo = 6;
			break;
		case DISP_STATUS3: // ステータス表示
			pTitle = &bm_menu_st;
			DrawBitmap(45, 70, &bm_menu_st3bk);
			nNo = 6;
			break;
		case DISP_MUTE: // ミュート
			pTitle = &bm_menu_mute;
			nNo = 7;
			break;
		}
		// タイトル
		DrawBitmap(29, 9, pTitle);
		
		// 左タブバー
		for(i = 0; i < 8; i++)
		{
			if (s_Current.Display.nMode > DISP_HOME)
			{
				// 再描画しなくても良い部分は飛ばす
				if (i != s_nLastNo && i != nNo)
					continue;
			}
			UINT32 col;
			if (i == 0) col = RGB(87, 155, 204);
			else if (i == nNo) col = RGB(219, 57, 57);
			else col = RGB(160, 160, 160);
			
			FillRect(6, i * 30 + 2, 13, 26, col);
		}
		
		s_nLastNo = nNo;
	}
	
	// 区切り線
	FillRect(30, 57, 358, 1, RGB(125, 229, 206));

	// Done
	if (nMode < DISP_STATUS1 || DISP_STATUS3 < nMode)
	{
		if (s_CtrlData.Display.bDone)
			DrawBitmap(323, 69, &bm_done);
		else
			ClearRect(323, 69, bm_done.w, bm_done.h);
	}
	
	// 設定内容
	int nOutput = s_CtrlData.Setting.nOutput;
	switch(nMode)
	{
	case DISP_OUTPUT: // 出力切替え
		pBm = (s_CtrlData.Setting.nOutput == 0)? &bm_menu_hpsp : &bm_menu_sphp;
		DrawBitmap(73, 133, pBm);
		break;
	
	case DISP_BBOOST: // バスブースト設定
		pBm = (s_CtrlData.Setting.nBBoost[nOutput] == 0)? &bm_menu_offon : &bm_menu_onoff;
		DrawBitmap(142, 133, pBm);
		break;

	case DISP_DACGAIN: // DACゲイン設定
		pBm = (s_CtrlData.Setting.nDacGain[nOutput] == 0)? &bm_menu_hilo : &bm_menu_lohi;
		DrawBitmap(133, 133, pBm);
		break;
		
	case DISP_PCMFLT1: // PCMフィルタ設定
	case DISP_PCMFLT2: // PCMフィルタ設定
	case DISP_PCMFLT3: // PCMフィルタ設定
	case DISP_PCMFLT4: // PCMフィルタ設定
	case DISP_PCMFLT5: // PCMフィルタ設定
		nFlt = s_CtrlData.Display.nMode - 5;
		nCur = s_CtrlData.Setting.nPcmFlt;
		
		if (nCur >= 5) --nCur; // 4と5は同じ対応
		if (nFlt >= nCur)
			nFlt++; // 現在値以外の値を設定する

		switch(nCur)
		{
		case 0: pBm = &bm_menu_shr; break;
		case 1: pBm = &bm_menu_slr; break;
		case 2: pBm = &bm_menu_shdshr; break;
		case 3: pBm = &bm_menu_shdslr; break;
		case 4: pBm = &bm_menu_sus; break;
		case 5: pBm = &bm_menu_lds; break;
		}
		x = 41;
		DrawBitmap(x, 120, pBm);
		x += pBm->w;
		ClearRect(x, 120, 380 - x, pBm->h);

		DrawBitmap(53, 149, &bm_menu_arrow);

		switch(nFlt)
		{
		case 0: pBm = &bm_menu_shr; break;
		case 1: pBm = &bm_menu_slr; break;
		case 2: pBm = &bm_menu_shdshr; break;
		case 3: pBm = &bm_menu_shdslr; break;
		case 4: pBm = &bm_menu_sus; break;
		case 5: pBm = &bm_menu_lds; break;
		}
		x = 76;
		DrawBitmap(x, 149, pBm);
		x += pBm->w;
		ClearRect(x, 149, 390 - x, pBm->h);
		break;
		
	case DISP_DSDFLT: // DSDフィルタ設定
		pBm = (s_CtrlData.Setting.nDsdFlt == 0)? &bm_menu_lohi : &bm_menu_hilo;
		DrawBitmap(133, 133, pBm);
		break;
		
	case DISP_STATUS1: // ステータス表示
	case DISP_STATUS2: // ステータス表示
	case DISP_STATUS3: // ステータス表示
		DrawStatus(nMode);
		break;
		
	case DISP_MUTE: // ミュート
		pBm = (s_CtrlData.Setting.nMute == 0)? &bm_menu_offon : &bm_menu_onoff;
		DrawBitmap(142, 133, pBm);
		break;
	}
}

// ●ステータス描画
static void DrawStatus(int nMode)
{
	int x, y, s;
	switch(nMode)
	{
	case DISP_STATUS1:
		x = 144; y = 96;
		DrawNumber(0, x, y, 0, s_CtrlData.Status.nState, DN_LEFT);
		y += 22;
		DrawNumber(0, x, y, 0, s_CtrlData.Status.nErrInfo, DN_LEFT);
		y += 22;
		DrawNumber(0, x, y, 0, s_CtrlData.Status.nVCXO, DN_LEFT);
		y += 22; s = x;
		x += DrawNumber(0, x, y, 0, s_CtrlData.Status.nAdvValue, DN_LEFT);
		if (s_CtrlData.Status.bAdvValid)
		{
			DrawBitmap(x, y, &bm_dig_w14_ast);
			x += bm_dig_w14_ast.w;
		}
		ClearRect(x, y, 195 - x, bm_dig_w14_0.h);
		y += 22; x = s;
		x += DrawNumber(0, x, y, 0, s_CtrlData.Setting.Avc[0].nAdcVal, DN_LEFT);
		ClearRect(x, y, 187 - x, bm_dig_w14_0.h);
		y += 22; x = s;
		x += DrawNumber(0, x, y, 0, s_CtrlData.Setting.Avc[1].nAdcVal, DN_LEFT);
		ClearRect(x, y, 187 - x, bm_dig_w14_0.h);

		x = 303; y = 96;
		DrawNumber(0, x + bm_dig_w14_0.w * 3, y, 0, (s_CtrlData.Status.nInputs & 0x01) >> 0, DN_LEFT);
		DrawNumber(0, x + bm_dig_w14_0.w * 2, y, 0, (s_CtrlData.Status.nInputs & 0x02) >> 1, DN_LEFT);
		DrawNumber(0, x + bm_dig_w14_0.w * 1, y, 0, (s_CtrlData.Status.nInputs & 0x04) >> 2, DN_LEFT);
		DrawNumber(0, x + bm_dig_w14_0.w * 0, y, 0, (s_CtrlData.Status.nInputs & 0x08) >> 3, DN_LEFT);
		y += 22;
		DrawNumber(0, x, y, 0, (s_CtrlData.Status.nInputs & COMBO384_DSDON_MASK) >> 6, DN_LEFT);
		y += 22;
		DrawNumber(0, x, y, 0, (s_CtrlData.Status.nInputs & COMBO384_DSD64_MASK) >> 5, DN_LEFT);
		y += 22;
		DrawNumber(0, x, y, 0, (s_CtrlData.Status.nInputs & COMBO384_MUTE_MASK) >> 4, DN_LEFT);
		y += 22; s = x;
		x += DrawNumber(0, x, y, 0, s_CtrlData.Status.nInputInfo, DN_LEFT);
		ClearRect(x, y, 324 - x, bm_dig_w14_0.h);
		y += 22; x = s;
		DrawNumber(0, x, y, 0, s_CtrlData.Status.nFormat, DN_LEFT);
		break;
		
	case DISP_STATUS2:
		x = 179; y = 96;
		DrawNumber(0, x, y, 0, s_CtrlData.Status.LMK04832.PLL1_LD, DN_LEFT);
		y += 22;
		DrawNumber(0, x, y, 0, s_CtrlData.Status.LMK04832.PLL2_LD, DN_LEFT);
		y += 22;
		DrawNumber(0, x, y, 0, s_CtrlData.Status.LMK04832.LOS, DN_LEFT);
		y += 22;
		DrawNumber(0, x, y, 0, s_CtrlData.Status.LMK04832.DAC_LOCKED, DN_LEFT);
		y += 22;
		x += DrawNumber(0, x, y, 0, s_CtrlData.Status.LMK04832.DAC_VALUE, DN_LEFT);
		ClearRect(x, y, 240 - x, bm_dig_w14_0.h);

		x = 355; y = 96;
		DrawNumber(0, x, y, 0, s_CtrlData.Status.LMK04832.PLL1_LD_LOST, DN_LEFT);
		y += 22;
		DrawNumber(0, x, y, 0, s_CtrlData.Status.LMK04832.PLL2_LD_LOST, DN_LEFT);
		y += 22;
		DrawNumber(0, x, y, 0, s_CtrlData.Status.LMK04832.HOLDOVER, DN_LEFT);
		y += 22;
		break;
		
	case DISP_STATUS3:
		x = 135; y = 121;
		DrawNumber(0, x, y, 0, s_CtrlData.Status.BiasCtrl[0].nStatus, DN_LEFT);
		y += 23; s = x;
		x += DrawNumber(0, x, y, 0, s_CtrlData.Status.BiasCtrl[0].nDacVal, DN_LEFT);
		ClearRect(x, y, 169 - x, bm_dig_w14_0.h);
		y += 23; x = s;
		x += DrawNumber(0, x, y, 0, s_CtrlData.Status.BiasCtrl[0].nAvgCur, DN_LEFT);
		ClearRect(x, y, 169 - x, bm_dig_w14_0.h);
		y += 23; x = s;
		x += DrawNumber(0, x, y, 0, s_CtrlData.Status.BiasCtrl[0].nSetCur, DN_LEFT);
		ClearRect(x, y, 169 - x, bm_dig_w14_0.h);

		x = 187; y = 121;
		DrawNumber(0, x, y, 0, s_CtrlData.Status.BiasCtrl[1].nStatus, DN_LEFT);
		y += 23; s = x;
		x += DrawNumber(0, x, y, 0, s_CtrlData.Status.BiasCtrl[1].nDacVal, DN_LEFT);
		ClearRect(x, y, 221 - x, bm_dig_w14_0.h);
		y += 23; x = s;
		x += DrawNumber(0, x, y, 0, s_CtrlData.Status.BiasCtrl[1].nAvgCur, DN_LEFT);
		ClearRect(x, y, 221 - x, bm_dig_w14_0.h);
		y += 23; x = s;
		x += DrawNumber(0, x, y, 0, s_CtrlData.Status.BiasCtrl[1].nSetCur, DN_LEFT);
		ClearRect(x, y, 221 - x, bm_dig_w14_0.h);

		x = 241; y = 121;
		DrawNumber(0, x, y, 0, s_CtrlData.Status.BiasCtrl[2].nStatus, DN_LEFT);
		y += 23; s = x;
		x += DrawNumber(0, x, y, 0, s_CtrlData.Status.BiasCtrl[2].nDacVal, DN_LEFT);
		ClearRect(x, y, 274 - x, bm_dig_w14_0.h);
		y += 23; x = s;
		x += DrawNumber(0, x, y, 0, s_CtrlData.Status.BiasCtrl[2].nAvgCur, DN_LEFT);
		ClearRect(x, y, 274 - x, bm_dig_w14_0.h);
		y += 23; x = s;
		x += DrawNumber(0, x, y, 0, s_CtrlData.Status.BiasCtrl[2].nSetCur, DN_LEFT);
		ClearRect(x, y, 274 - x, bm_dig_w14_0.h);

		x = 294; y = 121;
		DrawNumber(0, x, y, 0, s_CtrlData.Status.BiasCtrl[3].nStatus, DN_LEFT);
		y += 23; s = x;
		x += DrawNumber(0, x, y, 0, s_CtrlData.Status.BiasCtrl[3].nDacVal, DN_LEFT);
		ClearRect(x, y, 326 - x, bm_dig_w14_0.h);
		y += 23; x = s;
		x += DrawNumber(0, x, y, 0, s_CtrlData.Status.BiasCtrl[3].nAvgCur, DN_LEFT);
		ClearRect(x, y, 326 - x, bm_dig_w14_0.h);
		y += 23; x = s;
		x += DrawNumber(0, x, y, 0, s_CtrlData.Status.BiasCtrl[3].nSetCur, DN_LEFT);
		ClearRect(x, y, 326 - x, bm_dig_w14_0.h);
		break;
	}
}

// ●シャットダウン
static void Shutdown(void)
{
	// バックライト消灯
	LCD_ClrLED();

	// 点滅停止
	T2CONbits.ON = 0;
	
	// Display OFF Setting
	WriteCtrlReg(0x0007, 0x0000);
	WriteCtrlReg(0x0100, 0x0000);
	WriteCtrlReg(0x0102, 0x0000);
	WriteCtrlReg(0x0103, 0x0000);
	
	// 電源断まで無限ループ
	ClearBusy();
	for(;;) ClrWdt();	
}

// ●数値を描画する
static int DrawNumber(int nFont, int x, int y, int w, int nVal, int nAdj)
{
	// 白 size 14
	static SBitmap* s_Font1[] = { &bm_dig_w14_0,
		&bm_dig_w14_1, &bm_dig_w14_2, &bm_dig_w14_3, 
		&bm_dig_w14_4, &bm_dig_w14_5, &bm_dig_w14_6, 
		&bm_dig_w14_7, &bm_dig_w14_8, &bm_dig_w14_9 
	};
	// 緑 size 14
	static SBitmap* s_Font2[] = { &bm_dig_g14_0,
		&bm_dig_g14_1, &bm_dig_g14_2, &bm_dig_g14_3, 
		&bm_dig_g14_4, &bm_dig_g14_5, &bm_dig_g14_6, 
		&bm_dig_g14_7, &bm_dig_g14_8, &bm_dig_g14_9 
	};
	// 白 size 16
	static SBitmap* s_Font3[] = { &bm_dig_w16_0,
		&bm_dig_w16_1, &bm_dig_w16_2, &bm_dig_w16_3, 
		&bm_dig_w16_4, &bm_dig_w16_5, &bm_dig_w16_6, 
		&bm_dig_w16_7, &bm_dig_w16_8, &bm_dig_w16_9 
	};
	// デジタル size 74
	static SBitmap* s_Font4[] = { &bm_dig_d74_0,
		&bm_dig_d74_1, &bm_dig_d74_2, &bm_dig_d74_3, 
		&bm_dig_d74_4, &bm_dig_d74_5, &bm_dig_d74_6, 
		&bm_dig_d74_7, &bm_dig_d74_8, &bm_dig_d74_9 
	};
	static SBitmap** s_pFonts[] = {
		s_Font1, s_Font2, s_Font3, s_Font4
	};
	
	// 桁数カウント
	int nTemp = nVal;
	int nNum = 1;
	int nDiv = 1;
	for(;; nDiv *= 10, nNum++)
	{
		nTemp /= 10;
		if (nTemp == 0)
			break;
	}

	// 幅と開始位置算出
	SBitmap** pSet = s_pFonts[nFont];
	int nWidth = (pSet[0]->w) * nNum;
	
	int nLeft;
	if (nAdj == DN_CENTER) nLeft = x + (w - nWidth) / 2;
	else if (nAdj == DN_RIGHT) nLeft = x + w - nWidth;
	else nLeft = x;
	
	// 描画
	for(;; nDiv /= 10, nLeft += pSet[0]->w)
	{
		DrawBitmap(nLeft, y, pSet[nVal / nDiv]);
		if (--nNum == 0)
			break;
		
		nVal = nVal % nDiv;
	}
	
	return nWidth;
}

//---------------------------------------------------------------

// ●LCD(M032C9326D8)のリセット
static void ResetLCD(void)
{
	// リセット
	LCD_ClrRST();
	DelayMs(1);
	LCD_SetRST();
}

// ●LCDの初期化
static void InitLCD(void)
{
	// Display OFF Setting
	WriteCtrlReg(0x0007, 0x0000);
	WriteCtrlReg(0x0100, 0x0000);
	WriteCtrlReg(0x0102, 0x0000);
	WriteCtrlReg(0x0103, 0x0000);
	
	// Dis-charge capacitor power voltage
	DelayMs(200);

	// Power supply initial setting 
	WriteCtrlReg(0x0101, 0x0227); // DC1[2:0], DC0[2:0], VC[2:0]
	WriteCtrlReg(0x0102, 0x01AD); // VRH, VREG1OUT voltage
	WriteCtrlReg(0x0103, 0x0D00); // VDV[4:0] for VCOM amplitude
	WriteCtrlReg(0x0281, 0x000E); // VCM[5:0] for VCOMH

	// Power supply operation setting (1)
	WriteCtrlReg(0x0100, 0x0090); // APE, AP, DSTB, SLP, STB
	WriteCtrlReg(0x0102, 0x01BD); // PON

	DelayMs(40);

	// Power supply operation setting (2)
	WriteCtrlReg(0x0100, 0x1190); // BT[2:0]
	WriteCtrlReg(0x0103, 0x2D00); // VCOMG
	
	// その他の設定
	WriteCtrlReg(0x0001, 0x0000); // set SS and SM bit
	WriteCtrlReg(0x0002, 0x0100); // set 1 line inversion
	WriteCtrlReg(0x0003, 0xD038); // set TRI, DFM, BGR, I/D[1:0]=11.
	WriteCtrlReg(0x0008, 0x0202); // set the back porch and front porch
	WriteCtrlReg(0x0009, 0x0000); // set non-display area refresh cycle ISC[3:0]
	WriteCtrlReg(0x000C, 0x0000); // RGB interface setting
	WriteCtrlReg(0x000F, 0x0000); // RGB interface polarity
	
	// Panel Interface Control
	WriteCtrlReg(0x0010, 0x0010); // DIVI[1:0], RTNI[4:0]
	WriteCtrlReg(0x0011, 0x0600); // NOWI[2:0], SDTI[2:0]
	WriteCtrlReg(0x0020, 0x0002); // DIVE[1:0], RTNE[5:0]

	// Gamma Control 1～10
	WriteCtrlReg(0x0300, 0x0000);
	WriteCtrlReg(0x0301, 0x0707);
	WriteCtrlReg(0x0302, 0x0606);
	WriteCtrlReg(0x0305, 0x0000);
	WriteCtrlReg(0x0306, 0x0D00);
	WriteCtrlReg(0x0307, 0x0706);
	WriteCtrlReg(0x0308, 0x0005);
	WriteCtrlReg(0x0309, 0x0007);
	WriteCtrlReg(0x030C, 0x0000);
	WriteCtrlReg(0x030D, 0x000A);
	
	// Base Image Display Control
	WriteCtrlReg(0x0400, 0x3100); // Gate Scan Line 400 lines
	WriteCtrlReg(0x0401, 0x0001); // NDL,VLE, REV
	WriteCtrlReg(0x0404, 0x0000); // set scrolling line

	// Partial Display Control
	WriteCtrlReg(0x0500, 0x0000);
	WriteCtrlReg(0x0501, 0x0000);
	WriteCtrlReg(0x0502, 0x0000);
	WriteCtrlReg(0x0503, 0x0000);
	WriteCtrlReg(0x0504, 0x0000);
	WriteCtrlReg(0x0505, 0x0000);
	
	// サンプルにある非公開？設定
	WriteCtrlReg(0x0702, 0x3008); // Set internal timing, don’t change this value
	WriteCtrlReg(0x0705, 0x0036); // Set internal timing, don’t change this value
	WriteCtrlReg(0x070B, 0x1213); // Set internal timing, don’t change this value

	// 背景クリア
	ClearRect(0, 0, LCD_WIDTH, LCD_HEIGHT);
	
	// Display ON Sequence
	WriteCtrlReg(0x0100, 0x1090); // SAP

	WriteCtrlReg(0x0007, 0x0161); // VON=1, GON=1, DTE=0, D[1:0]=01
	DelayMs(40); // Wait for 2 frames or more
	WriteCtrlReg(0x0007, 0x0163); // VON=1, GON=1, DTE=0, D[1:0]=11
	DelayMs(40); // Wait for 2 frames or more
	WriteCtrlReg(0x0007, 0x0173); // VON=1, GON=1, DTE=1, D[1:0]=11
	DelayMs(40);

	// バックライト点灯
	LCD_SetLED();
}

// ●矩形の塗りつぶし
static void FillRect(int x, int y, int w, int h, UINT32 color)
{
	if (w <= 0 || h <= 0)
		return;
	
	BYTE r = GET_R(color);
	BYTE g = GET_G(color);
	BYTE b = GET_B(color);

	// GRAM Horizontal
	WriteCtrlReg(0x0200, y);

	// GRAM Vertical Address
	WriteCtrlReg(0x0201, x);

	// Horizontal and Vertical RAM Address Position
	WriteCtrlReg(0x0210, y);
	WriteCtrlReg(0x0211, y + h - 1);
	WriteCtrlReg(0x0212, x);
	WriteCtrlReg(0x0213, x + w - 1);

	LCD_ClrCS();

	// Write Data to GRAM 
	WriteReg(TRUE, 0x0202);

	LCD_SetRS();
	LCD_DirOutput();

	int count = w * h, i;
	for(i = 0; i < count; i++)
	{
		// 1st R
		LCD_ClrWR();
		LCD_DATA_O = r;
		DelayNs(20); // 50ns

		LCD_SetWR();
		DelayNs(20); // 50ns

		// 2nd G
		LCD_ClrWR();
		LCD_DATA_O = g;
		DelayNs(20); // 50ns

		LCD_SetWR();
		DelayNs(20); // 50ns

		// 3rd B
		LCD_ClrWR();
		LCD_DATA_O = b;
		DelayNs(20); // 50ns

		LCD_SetWR();
		DelayNs(20); // 50ns
	}

	LCD_SetCS();
}

// ●ビットマップの描画
static void DrawBitmap(int x, int y, SBitmap* pbm)
{
	// 負数の座標はセンタリング
	if (x < 0)
		x = (LCD_WIDTH - pbm->w) / 2;
	if (y < 0)
		y = (LCD_HEIGHT - pbm->h) / 2;
	
	// GRAM Horizontal
	WriteCtrlReg(0x0200, y);

	// GRAM Vertical Address
	WriteCtrlReg(0x0201, x);

	// Horizontal and Vertical RAM Address Position
	WriteCtrlReg(0x0210, y);
	WriteCtrlReg(0x0211, y + pbm->h - 1);
	WriteCtrlReg(0x0212, x);
	WriteCtrlReg(0x0213, x + pbm->w - 1);

	LCD_ClrCS();

	// Write Data to GRAM 
	WriteReg(TRUE, 0x0202);

	LCD_SetRS();
	LCD_DirOutput();

	const BYTE* p = pbm->p;
	int count = pbm->w * pbm->h, i;
	for(i = 0; i < count; i++)
	{
		// 1st R
		LCD_ClrWR();
		LCD_DATA_O = *p++;
		DelayNs(10); // 50ns

		LCD_SetWR();
		DelayNs(20); // 50ns

		// 2nd G
		LCD_ClrWR();
		LCD_DATA_O = *p++;
		DelayNs(10); // 50ns

		LCD_SetWR();
		DelayNs(20); // 50ns

		// 3rd B
		LCD_ClrWR();
		LCD_DATA_O = *p++;
		DelayNs(10); // 50ns

		LCD_SetWR();
		DelayNs(20); // 50ns
	}

	LCD_SetCS();
}

// ●DeviceIDを読み込む
static UINT16 ReadDeviceID(void)
{
	UINT16 nDeviceID;
	
	LCD_ClrCS();
	nDeviceID = ReadReg(TRUE);
	LCD_SetCS();
	
	return nDeviceID;
}

// ●制御レジスタに書き込む
static void WriteCtrlReg(UINT16 addr, UINT16 data)
{
	LCD_ClrCS();

	WriteReg(TRUE, addr);
	WriteReg(FALSE, data);

	LCD_SetCS();
}

// ●Register から読み込む
static UINT16 ReadReg(BOOL index)
{
	UINT16 data;
	
	if (index) LCD_ClrRS();
	else LCD_SetRS();

	DelayNs(5);
	LCD_DirInput();
	
	// High Byte
	LCD_ClrRD();
	DelayNs(80); // 150ns
	data = ((BYTE)LCD_DATA_I) << 8;
		
	LCD_SetRD();
	DelayNs(90); // 150ns

	// Low Byte
	LCD_ClrRD();
	DelayNs(80); // 150ns
	data |= (BYTE)LCD_DATA_I;
		
	LCD_SetRD();
	DelayNs(90); // 150ns
	
	return data;
}

// ●Control Register に書き込む
static void WriteReg(BOOL index, UINT16 data)
{
	if (index) LCD_ClrRS();
	else LCD_SetRS();

	DelayNs(10);
	LCD_DirOutput();
	
	// High Byte
	LCD_ClrWR();
	LCD_DATA_O = (BYTE)(data >> 8);
	DelayNs(20); // 50ns
		
	LCD_SetWR();
	DelayNs(20); // 50ns

	// Low Byte
	LCD_ClrWR();
	LCD_DATA_O = (BYTE)(data);
	DelayNs(20); // 50ns
		
	LCD_SetWR();
	DelayNs(20); // 50ns
}

//---------------------------------------------------------------

// ●NS待ち
static void Delay(register unsigned long time)
{
	// Counterリセットして指定の値になるまでループ
	asm("mtc0 $zero, $9, 0");
	while (__builtin_mfc0(9, 0) < time);
}

//---------------------------------------------------------------

// ●PORTD Input Change Interrupt
void __attribute__((vector(_CHANGE_NOTICE_D_VECTOR), interrupt(IPL6AUTO), nomips16))
CNDISR(void)
{
	volatile static SDispCtrl Data;
	volatile static UINT8* e = Data.Bytes + sizeof(Data.Bytes);

	if (CNFDbits.CNFD10)
	{
		volatile UINT8* p = Data.Bytes;
		do
		{
			*p = 0;
			while(COM_SCL);
			while(!COM_SCL);
			*p |= COM_SDA << 0;
			while(COM_SCL);
			while(!COM_SCL);
			*p |= COM_SDA << 1;
			while(COM_SCL);
			while(!COM_SCL);
			*p |= COM_SDA << 2;
			while(COM_SCL);
			while(!COM_SCL);
			*p |= COM_SDA << 3;
			while(COM_SCL);
			while(!COM_SCL);
			*p |= COM_SDA << 4;
			while(COM_SCL);
			while(!COM_SCL);
			*p |= COM_SDA << 5;
			while(COM_SCL);
			while(!COM_SCL);
			*p |= COM_SDA << 6;
			while(COM_SCL);
			while(!COM_SCL);
			*p |= COM_SDA << 7;
		}
		while(++p < e);
		
		// LOを出力→Busy状態にする
		COM_SCL_TRIS = 0;

		// マーカーチェク
		if (Data.nEndMarker == 0xA5)
		{
			// 想定通り→正式構造体へコピー
			s_CtrlData = Data;
		}
		else
		{
			COM_SCL_TRIS = 1;
			DEBUG_LED = 0;
		}
		
		CNFDbits.CNFD10 = 0;
	}
	
	IFS3bits.CNDIF = 0;
}

// Timer2/3
void __attribute__((vector(_TIMER_3_VECTOR), interrupt(IPL4AUTO), nomips16))
T3ISR(void)
{
	if (s_bSens)
		ClearRect(295, 157, bm_auto.w, bm_auto.h);
	else
		DrawBitmap(295, 157, &bm_sens);
	
	s_bSens = !s_bSens;

	IFS0bits.T3IF = 0;
}