/****************************************************************************

 ■HiResAmp 全体制御(MainControl)
		Copyright(C) 2020 電子うさぎ https://電子うさぎ.com/

****************************************************************************/

// PIC24F32KA304 Configuration Bit Settings

// FBS
#pragma config BWRP = OFF               // Boot Segment Write Protect (Disabled)
#pragma config BSS = OFF                // Boot segment Protect (No boot program flash segment)

// FGS
#pragma config GWRP = OFF               // General Segment Write Protect (General segment may be written)
#pragma config GSS0 = OFF               // General Segment Code Protect (No Protection)

// FOSCSEL
#pragma config FNOSC = PRI              // Oscillator Select (Primary Oscillator (XT, HS, EC))
#pragma config SOSCSRC = DIG            // SOSC Source Type (Digital Mode for use with external source)
#pragma config LPRCSEL = HP             // LPRC Oscillator Power and Accuracy (High Power, High Accuracy Mode)
#pragma config IESO = ON                // Internal External Switch Over bit (Internal External Switchover mode enabled (Two-speed Start-up enabled))

// FOSC
#pragma config POSCMOD = HS             // Primary Oscillator Configuration bits (HS oscillator mode selected)
#pragma config OSCIOFNC = OFF           // CLKO Enable Configuration bit (CLKO output disabled)
#pragma config POSCFREQ = MS            // Primary Oscillator Frequency Range Configuration bits (Primary oscillator/external clock input frequency between 100kHz and 8MHz)
#pragma config SOSCSEL = SOSCHP         // SOSC Power Selection Configuration bits (Secondary Oscillator configured for high-power operation)
#pragma config FCKSM = CSDCMD           // Clock Switching and Monitor Selection (Both Clock Switching and Fail-safe Clock Monitor are disabled)

// FWDT
#pragma config WDTPS = PS8192           // Watchdog Timer Postscale Select bits (1:8192)
#pragma config FWPSA = PR32             // WDT Prescaler bit (WDT prescaler ratio of 1:32)
#pragma config FWDTEN = SWON            // Watchdog Timer Enable bits (WDT controlled with the SWDTEN bit setting)
#pragma config WINDIS = OFF             // Windowed Watchdog Timer Disable bit (Standard WDT selected(windowed WDT disabled))

// FPOR
#pragma config BOREN = BOR3             // Brown-out Reset Enable bits (Brown-out Reset enabled in hardware, SBOREN bit disabled)
#pragma config LVRCFG = OFF             // Low Voltage Regulator Configuration bit (Low Voltage regulator is not available)
#pragma config PWRTEN = ON              // Power-up Timer Enable bit (PWRT enabled)
#pragma config I2C1SEL = SEC            // Alternate I2C1 Pin Mapping bit (Use  Alternate SCL1/ASDA1 Pins For I2C1)
#pragma config BORV = V18               // Brown-out Reset Voltage bits (Brown-out Reset set to lowest voltage (1.8V))
#pragma config MCLRE = ON               // MCLR Pin Enable bit (RA5 input pin disabled,MCLR pin enabled)

// FICD
#pragma config ICS = PGx1               // ICD Pin Placement Select bits (EMUC/EMUD share PGC1/PGD1)

// FDS
#pragma config DSWDTPS = DSWDTPSF       // Deep Sleep Watchdog Timer Postscale Select bits (1:2,147,483,648 (25.7 Days))
#pragma config DSWDTOSC = LPRC          // DSWDT Reference Clock Select bit (DSWDT uses Low Power RC Oscillator (LPRC))
#pragma config DSBOREN = ON             // Deep Sleep Zero-Power BOR Enable bit (Deep Sleep BOR enabled in Deep Sleep)
#pragma config DSWDTEN = OFF            // Deep Sleep Watchdog Timer Enable bit (DSWDT disabled)

#include <xc.h>
#include <p24F32KA304.h>
#include <libpic30.h>

//=================================================================

#define I2C_AMPCTRL		0x51

#define DEBUG_LED		_LATB14
#define POWER_CTRL		_LATB4
#define POWER_LED		_LATA4
#define SOUNDER			_LATB8
#define EXTCTL			_LATA9

#define COM_CLK			_RB6
#define COM_DAT			_LATB5
#define COM_CLK_TRIS	TRISBbits.TRISB6
#define COM_DAT_TRIS	TRISBbits.TRISB5

#define RE1_A			_RB10
#define RE1_B			_RA6
#define RE1_E			_RA7
#define RE2_A			_RB11
#define RE2_B			_RB12
#define RE2_E			_RB7
#define IR_IN			_RC0

#define RE_NONE			0
#define RE_UP			1
#define RE_DOWN			2

#define CUSTOM_CODE		0x3B86 // リモコン固有のコード
#define MARGIN			0.5
#define TMAG			500 // 1ms/2us=500
#define READER1_S		((9.00 - 2) * TMAG)
#define READER1_L		((9.00 + MARGIN) * TMAG)
#define READER2_S		((4.50 - MARGIN) * TMAG)
#define READER2_L		((4.50 + MARGIN) * TMAG)
#define REPEAT_S		((2.25 - MARGIN) * TMAG)
#define REPEAT_L		((2.25 + MARGIN) * TMAG)
#define DATBIT_S		((0.56 - MARGIN) * TMAG)
#define DATBIT_L		((0.56 + MARGIN) * TMAG)
#define LODAT_S			((1.125 - MARGIN) * TMAG)
#define LODAT_L			((1.125 + MARGIN) * TMAG)
#define HIDAT_S			((2.25 - MARGIN) * TMAG)
#define HIDAT_L			((2.25 + MARGIN) * TMAG)
#define REP_COUNT		3
#define REPEAT			0x80

#define KEY_NONE		0x00
#define KEY_ENTER		0x01
#define KEY_BTN1		0x02
#define KEY_BTN2		0x03
#define KEY_VUP			0x04
#define KEY_VDN			0x05

// アンプ制御状態
#define STATE_INIT			0		// 初期化中(共通)
#define STATE_STANDBY		1		// 待機中
#define STATE_CONNECTED		2		// 接続済
#define STATE_PLAYING		3		// 再生中
#define STATE_ERROR			4		// エラー発生
#define STATE_SHUTDOWN		5		// シャットダウン(共通)

// アンプ制御コマンド
#define CMD_SET_VOLUME		1		// ボリューム設定
#define CMD_SET_BBOOST		2		// バスブースト設定
#define CMD_SET_OUTPUT		3		// 出力切り替え(SP/HP)
#define CMD_SET_MUTE		4		// ミュート
#define CMD_SET_GAIN		5		// DACゲイン設定
#define CMD_SET_FLTPCM		6		// DACフィルタ設定(PCM)
#define CMD_SET_FLTDSD		7		// DACフィルタ設定(DSD)
#define CMD_SET_AUTOVL		8		// オートボリューム設定
#define CMD_SHUTDOWN		99		// シャットダウン(共通)

// 表示モード
#define DISP_INIT			0	// 初期画面
#define DISP_HOME			1	// ホーム画面
#define DISP_OUTPUT			2	// 出力切替え
#define DISP_BBOOST			3	// バスブースト設定
#define DISP_DACGAIN		4	// DACゲイン設定
#define DISP_PCMFLT1		5	// PCMフィルタ設定
#define DISP_PCMFLT2		6	// PCMフィルタ設定
#define DISP_PCMFLT3		7	// PCMフィルタ設定
#define DISP_PCMFLT4		8	// PCMフィルタ設定
#define DISP_PCMFLT5		9	// PCMフィルタ設定
#define DISP_DSDFLT			10	// DSDフィルタ設定
#define DISP_STATUS1		11	// ステータス表示
#define DISP_STATUS2		12	// ステータス表示
#define DISP_STATUS3		13	// ステータス表示
#define DISP_MUTE			14	// ミュート
#define DISP_SHUTDOWN		99	// シャットダウン

// 外部PC管理状態
#define PC_NONE				0	// PC管理なし
#define PC_ENABLE			1	// 制御アプリ認識完了
#define PC_ERROR			2	// 制御アプリエラー

//---------------------------------------------------------------

typedef enum _BOOL { FALSE = 0, TRUE } BOOL;
typedef unsigned char UINT8;
typedef unsigned short int UINT16;
typedef unsigned long int UINT32;

#pragma pack(push)
#pragma pack(1)
// アンプ制御と共通↓↓↓↓↓
// LMK04832ステータス
typedef struct {
	UINT8 LOS			: 1;
	UINT8 PLL1_LD		: 1;
	UINT8 PLL2_LD		: 1;
	UINT8 PLL1_LD_LOST	: 1;
	UINT8 PLL2_LD_LOST	: 1;
	UINT8 HOLDOVER		: 1;
	UINT8 DAC_LOCKED	: 1;
	UINT16 DAC_VALUE;
} SLMK04832;

// バイアス制御ステータス
typedef struct {
	UINT8 nStatus;
	UINT8 nDacVal;
	UINT16 nAvgCur;
	UINT16 nSetCur;
} SBiasCtrl;

// ステータス総合
typedef struct {
	UINT8 bCmdBusy	: 1;	// コマンドBusy
	UINT8 nState	: 3;	// アンプ制御の状態
	UINT8 nErrInfo	: 4;	// エラー状態時の付加情報
	UINT8 nFormat	: 7;	// 再生フォーマット
	UINT8 nVCXO		: 1;	// VCXO周波数セレクト
	UINT8 nInputs;			// COMBO384 IO値
	UINT8 nInputInfo;		// 量子化ビット数又はDDSサンプリング周波数
	UINT8 nAnaATT;			// アナログアッテネータ
	UINT8 nDigATT;			// デジタルアッテネータ
	UINT8 bAdvValid;		// 音声平均電圧有効
	UINT16 nAdvAvg;			// 音声平均電圧
	SLMK04832 LMK04832;
	SBiasCtrl BiasCtrl[4];
} SStatus;

// 動作設定
typedef struct {
	UINT8 bEnable;			// 0or1
	UINT16 nVolume;			// 音量設定時Volume
	UINT16 nAdcVal;			// 音量設定値(ADC値)	
} SAutoVolCtrl;

typedef struct {
	UINT8 nVolume[2];		// HP=0～72, SP=0～81
	UINT8 nBBoost[2];		// 0or1
	UINT8 nOutput	: 1;	// 0:HP, 1:SP
	UINT8 nMute		: 1;	// 0or1
	UINT8 nPcmFlt	: 3;	// 0～6(4と5は同じ)
	UINT8 nDsdFlt	: 1;	// 0or1
	UINT8 nDacGain[2];		// 0or1
	SAutoVolCtrl Avc[2];
} SSetting;

// 受信データ
typedef union {
	struct {
		SStatus Status;
		SSetting Setting;
	};
	UINT8 Bytes[sizeof(SStatus) + sizeof(SSetting)];
} SAmpCtrl;
// アンプ制御と共通↑↑↑↑↑

// 表示制御と共通↓↓↓↓↓
// 表示設定
typedef struct {
	UINT8 nMode;		// 画面タイプ
	UINT8 bDone	:	1;	// Doneフラグ
	UINT8 nPCState:	4;	// 外部PC管理状態
} SDisplay;

// 送信データ
typedef union {
	struct {
		SStatus Status;
		SSetting Setting;
		SDisplay Display;
		UINT8 nEndMarker; 
	};
	UINT8 Bytes[sizeof(SStatus) + sizeof(SSetting) + sizeof(SDisplay) + 1];
} SDispCtrl;
// 表示制御と共通↑↑↑↑↑
#pragma pack(pop)

//---------------------------------------------------------------

static void InitSystem(void);
static void InitApp(void);
static void ClearApp(void);
static void Run(void);

static void PowerUp(void);
static void Home(void);
static BOOL Menu(void);
static void DoSet(void);
static BOOL PowerDown(void);
static void ClearInput(void);
static void AmpCtrl(UINT8 cmd, UINT8 arg, BOOL bWait);
static void DispCtrl(void);
static void ExtCtrl(const char* pCmd);
static BOOL CheckExtRes(const char* pRes);
static void Beep(UINT16 nFreq, UINT16 ms);
static int GetIR(void);
static UINT8 ToKey(UINT8 code);

static void SendI2C2(UINT8 nAddr, const UINT8* pData, int nCount);
static void RecvI2C2(UINT8 nAddr, UINT8* pData, int nCount);
static void SendComm(const UINT8* pData, int nCount);
static void SendUART(const char* pLine);
static UINT16 GetElapsed(UINT16 nBase);
static UINT16 GetElapsed2(UINT16 nBase);
static void DelayMs(UINT16 ms);

#define IsDispBusy()	(!COM_CLK)

//---------------------------------------------------------------

static volatile int s_bRE1A;
static volatile int s_bRE2A;
static volatile int s_nRE1;
static volatile int s_nRE2;
static volatile BOOL s_bSW1;
static volatile BOOL s_bSW2;
static volatile BOOL s_bRecved;
static volatile char s_Uart[64];
static volatile UINT16 s_nTick;
static volatile int s_nBeep;

static SDispCtrl s_CtrlData;
static SDispCtrl s_OldData;

//=================================================================

// ●メインエントリ
int main(void)
{
	InitSystem();
	InitApp();	
	Run();
	return 0;
}

// ●システム初期化
static void InitSystem(void)
{
	// ポート初期化
	PORTA = PORTB = 0x00;
	LATA = LATB =0x00;
	
	// デバッグLED
	TRISBbits.TRISB14 = 0;
	ANSELBbits.ANSB14 = 0;
	DEBUG_LED = 1;
	
	// POWER_CTRL
	TRISBbits.TRISB4 = 0;
	ANSELBbits.ANSB4 = 0;

	// POWER_LED
	TRISAbits.TRISA4 = 0;
	
	// SOUNDER
	TRISBbits.TRISB8 = 0;

	// EXTCTL
	TRISAbits.TRISA9 = 0;
	
	// RE1_A
	CNEN2bits.CN16IE = 1;

	// RE1_E
	CNPU1bits.CN9PUE = 1;
	CNEN1bits.CN9IE = 1;

	// RE2_A
	CNEN1bits.CN15IE = 1;

	// RE2_B
	ANSELBbits.ANSB12 = 0;

	// RE2_E
	CNPU2bits.CN23PUE = 1;
	CNEN2bits.CN23IE = 1;
	
	// IR_IN
	ANSELCbits.ANSC0 = 0;
	CNPU3bits.CN32PUE = 1;
	CNEN3bits.CN32IE = 1;
	
	// I2C2(AC1 Master)
	ANSELBbits.ANSB2 = 0;
	ANSELBbits.ANSB3 = 0;
	I2C2BRG = 4; // 400KHz at FCY=4MHz
	I2C2CONbits.DISSLW = 1; // ノイズ低減のため
	I2C2CONbits.I2CEN = 1;
	
	// UART1
	U1BRG = 25; // 9600
	U1MODEbits.UARTEN = 1;
	U1STAbits.UTXEN = 1;
	_U1RXIE = 1;
	
	// Timer1(Tick)
	T1CONbits.TCS = 0; // Internal clock (FCY)
	T1CONbits.TCKPS = 0b01; // 1:8 (2us, FCY=4MHz)
	PR1 = 500; // 1ms
	_T1IP = 3;
	_T1IE = 1;
	
	// Timer2(Buzzer)
	T2CONbits.TCS = 0; // Internal clock (FCY)
	T2CONbits.TCKPS = 0b00; // 1:1 (0.25us, FCY=4MHz)
	_T2IP = 6;
	_T2IE = 1;

	// Timer3(Debug LED)
	T3CONbits.TCS = 0; // Internal clock (FCY)
	T3CONbits.TCKPS = 0b11; // 1:256 (64us, FCY=4MHz)
	PR3 = 7812; // 500ms
	_T3IP = 1;
	_T3IE = 1;
	
	// Timer4(IR)
	T4CONbits.TCS = 0; // Internal clock (FCY)
	T4CONbits.TCKPS = 0b01; // 1:8 (2us, FCY=4MHz)
	T4CONbits.TON = 1;

	// Timer5(Power LED)
	T5CONbits.TCS = 0; // Internal clock (FCY)
	T5CONbits.TCKPS = 0b11; // 1:256 (64us, FCY=4MHz)
	PR5 = 7812; // 500ms
	_T5IP = 1;
	_T5IE = 1;

	// ChangeNotify
	_CNIP = 5;
	_CNIE = 1;
}

// ●アプリ初期化
static void InitApp(void)
{
	s_nTick = 0;
	T1CONbits.TON = 1;

	ClearApp();
}

// ●アプリデータ初期化
static void ClearApp(void)
{
	s_bRecved = FALSE;
	
	int i = 0;
	do s_CtrlData.Bytes[i] = 0;
	while(++i < sizeof(s_CtrlData.Bytes));
	s_CtrlData.nEndMarker = 0xA5;
	s_OldData = s_CtrlData;
}

// ●実行ループ
static void Run(void)
{
#ifndef _DEBUG
	_SWDTEN = 1;
#endif

	BOOL bPCDown = FALSE;
	UINT16 nBase = 0;
	int nCount = 0;
	
	DelayMs(500);
	ClearInput();
	
	// 電源断まで無限ループ
	for(;;)
	{
		// WDTタイムアウトか何か操作があるまでスリープ
		if (!bPCDown)
			Sleep();

		// リモコン受信
		int nRemocon = GetIR();
		
		// 電源ON操作チェック
		if (s_bSW1 || s_bSW2 || s_nRE1 != RE_NONE || s_nRE2 != RE_NONE || nRemocon == KEY_ENTER)
		{
			ClearApp();
			ClearInput();
			
			PowerUp();
			Home();

			bPCDown = PowerDown();
			nBase = s_nTick;
			nCount = 0;

			ClearInput();
		}
		
		// 通常の外部PCシャットダウンシーケンスで終わらなかった場合
		// 1分間の間シャットダウンコマンドを送り続ける
		if (bPCDown)
		{
			if (!CheckExtRes("OK") && nCount < 60)
			{
				// 外部PCにシャットダウンコマンドを送る
				if (GetElapsed(nBase) >= 1000)
				{
					SendUART("SHUTDOWN");
					nBase = s_nTick;
					nCount++;
				}
			}
			else
			{
				bPCDown = FALSE;
			}
		}
	}
}

// ●パワーアップ
static void PowerUp(void)
{
	Beep(4000, 100);
	
	// 電源ON
	POWER_LED = 1;
	POWER_CTRL = 1;
	
	// 外部PC起動
	EXTCTL = 1;
	DelayMs(250);
	EXTCTL = 0;
	
	// 安定待ち(アンプ制御のI2Cが有効になるまで待つ意味もある)
	DelayMs(250);
	
	// アンプ制御のスタンバイを確認する
	SAmpCtrl ctrl;
	for(;; DelayMs(100))
	{
		RecvI2C2(I2C_AMPCTRL, ctrl.Bytes, 1);
#ifndef _DEBUG
		if (ctrl.Bytes[0] == 0xFF)
			continue; // I2C通信できてない→WDTによるリセットに持ち込む
#endif		
		if (ctrl.Status.nState >= STATE_STANDBY)
			break;
	}

	// デバッグLED点滅
	DEBUG_LED = 0;
	TMR3 = 0;
	T3CONbits.TON = 1;
}

// ●通常動作(ホーム画面)
static void Home(void)
{
	BOOL bUpdate = TRUE;
	UINT16 nBase = 0;

	// ホーム画面に設定
	s_CtrlData.Display.nMode = 1;
	
	// OFF操作まで無限ループ
	do
	{
		ClrWdt();
		
		if (bUpdate || GetElapsed(nBase) >= 200)
		{
			// ステータス取得
			RecvI2C2(I2C_AMPCTRL, s_CtrlData.Bytes, sizeof(SAmpCtrl));
			nBase = s_nTick;

			// 表示情報送信
			DispCtrl();
		}
		if (bUpdate)
		{
			// 入力をリセット
			ClearInput();
			bUpdate = FALSE;
		}
		
		// リモコン受信
		int nRemocon = GetIR();

		// ボリュームノブ回転 →ボリューム UP/DOWN
		if (s_nRE1 == RE_UP || (nRemocon & ~REPEAT) == KEY_VUP)
		{
			AmpCtrl(CMD_SET_VOLUME, TRUE, FALSE);
			bUpdate = TRUE;
			continue;
		}
		if (s_nRE1 == RE_DOWN || (nRemocon & ~REPEAT) == KEY_VDN)
		{
			AmpCtrl(CMD_SET_VOLUME, FALSE, FALSE);
			bUpdate = TRUE;
			continue;
		}

		// ボリュームノブ押下 →オートボリューム ON/OFF
		if (s_bSW1)
		{
			AmpCtrl(CMD_SET_AUTOVL, !s_CtrlData.Setting.Avc[s_CtrlData.Setting.nOutput].bEnable, FALSE);
			bUpdate = TRUE;
			continue;
		}

		// リモコンボタン1(W) →アウトプット切替え・次のジャンル(外部PC接続時)
		if (nRemocon == KEY_BTN1)
		{
			if (s_CtrlData.Display.nPCState != PC_ENABLE)
				AmpCtrl(CMD_SET_OUTPUT, !s_CtrlData.Setting.nOutput, FALSE);
			else
			{
				ExtCtrl("NEXT2");
				Beep(3500, 50);
				DelayMs(100);
				Beep(3500, 50);
			}
			
			DelayMs(500); // 連続防止
			bUpdate = TRUE;
			continue;
		}

		// リモコンボタン2(T) →ミュート・次のプレイリスト(外部PC接続時)
		if (nRemocon == KEY_BTN2)
		{
			if (s_CtrlData.Display.nPCState != PC_ENABLE)
				AmpCtrl(CMD_SET_MUTE, !s_CtrlData.Setting.nMute, TRUE);
			else
			{
				ExtCtrl("NEXT1");
				Beep(3500, 50);
				DelayMs(100);
				Beep(3500, 50);
			}
			
			DelayMs(500); // 連続防止
			bUpdate = TRUE;
			continue;
		}

		// リモコンEnter →電源OFF
		if (nRemocon == KEY_ENTER)
			break;

		// ファンクション回転 →設定メニュー切替え
		if (s_nRE2 != RE_NONE)
		{
			if (Menu())
			{
				bUpdate = TRUE;
				continue;
			}
			break; // 電源OFF選択
		}
		
		// 外部PCチェック
		if (s_bRecved)
		{
			_U1RXIE = 0;
			if (CheckExtRes("STARTED"))
			{
				s_CtrlData.Display.nPCState = PC_ENABLE;
			}
			else if (CheckExtRes("ERR"))
			{
				s_CtrlData.Display.nPCState = PC_ERROR;
			}
			s_bRecved = FALSE;
			_U1RXIE = 1;
		}
	}
	while(!s_bSW2); // ファンクション押下 →電源OFF
}

// ●通常動作(メニュー画面)
static BOOL Menu(void)
{
	UINT16 nClose = s_nTick;
	UINT16 nStart = s_nTick;
	
	// 終了までループ
	for(;;)
	{
		ClrWdt();

		// ファンクション回転 →設定メニュー切替え
		if (s_nRE2 == RE_UP)
		{
			if (!s_CtrlData.Display.bDone)
			{
				if (++s_CtrlData.Display.nMode > DISP_MUTE)
					s_CtrlData.Display.nMode = DISP_HOME;
			}
			else
			{
				s_CtrlData.Display.bDone = FALSE;
			}
		}
		else if (s_nRE2 == RE_DOWN)
		{
			if (!s_CtrlData.Display.bDone)
			{
				if (--s_CtrlData.Display.nMode < DISP_HOME)
					s_CtrlData.Display.nMode = DISP_MUTE;
			}
			else
			{
				s_CtrlData.Display.bDone = FALSE;
			}
		}
		// 画面変更
		if (s_nRE2 != RE_NONE)
		{
			// ホームに戻る
			if (s_CtrlData.Display.nMode == 1)
				break;

			// 最新の設定を取得
			RecvI2C2(I2C_AMPCTRL, s_CtrlData.Bytes, sizeof(SAmpCtrl));

			// 表示情報送信
			DispCtrl();
			
			nStart = s_nTick;
			nClose = s_nTick;
			
			ClearInput();
			continue;
		}
		
		if ((s_CtrlData.Display.nMode < DISP_STATUS1) || (DISP_STATUS3 < s_CtrlData.Display.nMode))
		{
			// ファンクション押下 →設定実行
			if (s_bSW2)
			{
				if (!s_CtrlData.Display.bDone)
				{
					DoSet();
					s_CtrlData.Display.bDone = TRUE;

					// 表示情報送信
					DispCtrl();
				}

				nClose = s_nTick;
				ClearInput();
			}
		}
		else
		{
			// ファンクション押下 →抜ける
			if (s_bSW2)
				break;

			// ステータス表示更新
			if (GetElapsed(nStart) >= 200)
			{
				// ステータス取得
				RecvI2C2(I2C_AMPCTRL, s_CtrlData.Bytes, sizeof(SAmpCtrl));

				// 表示情報送信
				DispCtrl();

				nStart = s_nTick;
			}
		}
		
		// リモコン受信
		int nRemocon = GetIR();

		// ボリュームノブ回転 →ボリューム UP/DOWN
		if (s_nRE1 == RE_UP || (nRemocon & ~REPEAT) == KEY_VUP)
		{
			AmpCtrl(CMD_SET_VOLUME, TRUE, FALSE);
			ClearInput();
		}
		if (s_nRE1 == RE_DOWN || (nRemocon & ~REPEAT) == KEY_VDN)
		{
			AmpCtrl(CMD_SET_VOLUME, FALSE, FALSE);
			ClearInput();
		}
		
		// リモコンEnter →電源OF
		if (nRemocon == KEY_ENTER)
			return FALSE;

		// 無操作クローズ
		if ((s_CtrlData.Display.nMode < DISP_STATUS1) || (DISP_STATUS3 < s_CtrlData.Display.nMode))
		{
			if (GetElapsed(nClose) >= 6000)
				break;
		}
	}

	s_CtrlData.Display.nMode = 1;
	s_CtrlData.Display.bDone = FALSE;

	ClearInput();
	
	return TRUE;
}

// ●設定実行
static void DoSet(void)
{
	int nIndex = s_CtrlData.Setting.nOutput;
	UINT8 nCur, nFlt; 
	
	switch(s_CtrlData.Display.nMode)
	{
	case DISP_OUTPUT: // 出力切り替え
		AmpCtrl(CMD_SET_OUTPUT, !s_CtrlData.Setting.nOutput, TRUE);
		break;
		
	case DISP_MUTE: // ミュート
		AmpCtrl(CMD_SET_MUTE, !s_CtrlData.Setting.nMute, TRUE);
		break;

	case DISP_BBOOST: // バスブースト設定
		AmpCtrl(CMD_SET_BBOOST, !s_CtrlData.Setting.nBBoost[nIndex], TRUE);
		break;
		
	case DISP_DACGAIN: // DACゲイン設定
		AmpCtrl(CMD_SET_GAIN, !s_CtrlData.Setting.nDacGain[nIndex], TRUE);
		break;
		
	case DISP_PCMFLT1: case DISP_PCMFLT2: case DISP_PCMFLT3:
	case DISP_PCMFLT4: case DISP_PCMFLT5:
		// PCMフィルタ設定
		nFlt = s_CtrlData.Display.nMode - 5;
		nCur = s_CtrlData.Setting.nPcmFlt;
		
		if (nCur >= 5) --nCur; // 4と5は同じ対応
		if (nFlt >= nCur)
			nFlt++; // 現在値以外の値を設定する
		if (nFlt == 5) nFlt = 6; // 4と5は同じ対応
		
		AmpCtrl(CMD_SET_FLTPCM, nFlt, TRUE);
		break;
		
	case DISP_DSDFLT: // DSDフィルタ設定
		AmpCtrl(CMD_SET_FLTDSD, !s_CtrlData.Setting.nDsdFlt, TRUE);
		break;
	}
}

// ●パワーダウン
static BOOL PowerDown(void)
{
	Beep(500, 50);

	// パワーLED点滅開始
	POWER_LED = 0;
	T5CONbits.TON = 1;

	// アンプ制御へシャットダウン
	AmpCtrl(CMD_SHUTDOWN, 0, FALSE);

	// 表示制御へシャットダウン
	s_CtrlData.Display.nMode = DISP_SHUTDOWN;
	DispCtrl();

	// 外部PCシャットダウン
	ExtCtrl("SHUTDOWN");

	// アンプ制御シャットダウン待ち
	for(;; DelayMs(100))
	{
		SAmpCtrl ctrl;
		RecvI2C2(I2C_AMPCTRL, ctrl.Bytes, 1);
		if (ctrl.Status.nState == STATE_SHUTDOWN && !ctrl.Status.bCmdBusy)
			break;
	}
	// 表示制御シャットダウン待ち
	while(IsDispBusy());
	
	// 電源OFF
	POWER_CTRL = 0;
	
	// 電圧が落ちるまで待つ(PC_NONEの場合、1秒の応答待ちをしている)
	int nDelay = (s_CtrlData.Display.nPCState != PC_NONE)? 2000 : 1000;
	DelayMs(nDelay);
	
	// 外部PCシャットダウン待ち
	BOOL bPCDown = TRUE;
	if (s_CtrlData.Display.nPCState != PC_NONE)
	{
		DelayMs(30000);
		bPCDown = FALSE;
	}

	// パワーLED点滅終了
	T5CONbits.TON = 0;
	POWER_LED = 0;

	// デバッグLED点灯
	T3CONbits.TON = 0;
	DEBUG_LED = 1;
	
	return bPCDown;
}

// ●入力をクリアする
static void ClearInput(void)
{
	// ロータリーエンコーダの押し込みが解除されるまで待つ
	UINT16 i = 0;
	do {
		if (!RE1_E || !RE2_E)
			i = 0;
	}
	while (++i < 10000U);

	s_nRE1 = RE_NONE;
	s_nRE2 = RE_NONE;
	s_bSW1 = FALSE;
	s_bSW2 = FALSE;
	s_bRE1A = RE1_A;
	s_bRE2A = RE2_A;
}

// ●アンプ制御部へコマンドを出す
static void AmpCtrl(UINT8 cmd, UINT8 arg, BOOL bWait)
{
	// Busy解除待ち
	SAmpCtrl ctrl;
	for(;; DelayMs(10))
	{
		RecvI2C2(I2C_AMPCTRL, ctrl.Bytes, 1);
		if (!ctrl.Status.bCmdBusy)
			break;
	}
	// コマンド発行
	UINT8 data[2] = { cmd, arg };
	SendI2C2(I2C_AMPCTRL, data, 2);
	
	if (bWait) // 完了待ち
	{
		do {
			DelayMs(10);
			RecvI2C2(I2C_AMPCTRL, ctrl.Bytes, 1);
		}
		while (ctrl.Status.bCmdBusy);
	}
}

// ●表示制御部へ制御情報を出す
static void DispCtrl(void)
{
	// 最後に送信したデータと比較する
	int i = 0;
	do {
		if (s_CtrlData.Bytes[i] != s_OldData.Bytes[i])
			break;
	}
	while(++i < sizeof(s_CtrlData.Bytes));

	// 違う時だけ送信
	if (i != sizeof(s_CtrlData.Bytes))
	{
		// 制御情報送信
		SendComm(s_CtrlData.Bytes, sizeof(s_CtrlData.Bytes));
		s_OldData = s_CtrlData;
	}
}

// ●外部PCへコマンドを出す
static void ExtCtrl(const char* pCmd)
{
	// 念のためゴミクリア
	while(_U1RXIF);
	s_bRecved = FALSE;
	*s_Uart = '\0';
			
	SendUART(pCmd);
	
	// 応答待ち
	UINT16 nBase = s_nTick;
#ifdef _DEBUG
	while(!s_bRecved && GetElapsed(nBase) < 5000);
#else
	while(!s_bRecved && GetElapsed(nBase) < 1000);
#endif
	
	_U1RXIE = 0;
	if(!CheckExtRes("OK"))
	{
		// 期待する応答がなかった →外部PCはなくなったとする
		s_CtrlData.Display.nPCState = PC_NONE;
	}
	s_bRecved = FALSE;
	_U1RXIE = 1;
}

// ●外部PCからの応答をチェックする
static BOOL CheckExtRes(const char* pRes)
{
	if (s_bRecved)
	{
		int i;
		for(i = 0; ; i++)
		{
			char c = s_Uart[i];

			if (c != pRes[i])
				break;

			if (c == '\0')
				return TRUE;
		}
	}

	return FALSE;
}

// ●ビープ
static void Beep(UINT16 nFreq, UINT16 ms)
{
	UINT32 temp = ms * (UINT32)nFreq;
	s_nBeep = temp / 500UL;

	PR2 = 2000000 / nFreq;
	TMR2 = 0;
	T2CONbits.TON = 1;
}

// ●リモコンキー取得
static int GetIR(void)
{
#define IR_HI   (!IR_IN)
#define IR_LO	(IR_IN)
#define IR_TMR	TMR4	

	static UINT8 s_nRem = KEY_NONE;
	static UINT8 s_nState = 0;
	static UINT16 s_nStart;
	static UINT16 s_wCst;
	static UINT8 s_bDat;
	static UINT8 s_nDat;
	static UINT8 s_nBit;
	static int s_nRep;

	int nKey = KEY_NONE;
	do
	{
		// IR検出
		switch (s_nState)
		{
		case 0:
			// リーダーコード開始検出
			if (IR_HI)
			{
				s_nStart = IR_TMR;
				s_nState++;
			}
			break;

		case 1:
			// リーダーコード終了検出1
			if (IR_LO)
			{
				if (GetElapsed2(s_nStart) >= READER1_S)
				{
					s_nStart = IR_TMR;
					s_nState++;
				}
				else {
					s_nState = 0;
				}
			}
			else
			{
				if (GetElapsed2(s_nStart) > READER1_L) {
					s_nState = 0;
				}
			}
			break;

		case 2:
			// リーダーコード終了検出2
			if (IR_HI)
			{
				UINT16 w = GetElapsed2(s_nStart);
				if ((REPEAT_S <= w) && (w <= REPEAT_L))
				{
					// リピートコード検出!
					if (s_nRem != KEY_NONE)
					{
						if (s_nRep >= REP_COUNT) {
							nKey = s_nRem | REPEAT;
						}
						s_nRep++;
					}
				}
				else if (w >= READER2_S)
				{
					// リーダーコードを認識！
					s_wCst = s_bDat = s_nDat = s_nBit = 0;
					s_nRem = KEY_NONE;
					s_nStart = IR_TMR;
					s_nState++;
					break;
				}
				s_nState = 0;
			}
			else
			{
				if (GetElapsed2(s_nStart) > READER2_L)
				{
					s_nRem = KEY_NONE;
					s_nState = 0;
				}
			}
			break;

		case 3:
			// ビット検出1
			if (IR_LO)
			{
				if (GetElapsed2(s_nStart) >= DATBIT_S) {
					s_nState++;
				}
			}
			else
			{
				if (GetElapsed2(s_nStart) > DATBIT_L) {
					s_nState = 0;
				}
			}
			break;

		case 4:
			// ビット検出2
			if (IR_HI)
			{
				UINT8 b;
				UINT16 w = GetElapsed2(s_nStart);
				if ((LODAT_S <= w) && (w <= LODAT_L))
				{
					// '0'検出!
					b = 0x00;
				}
				else if (w <= HIDAT_L)
				{
					// '1'検出!
					b = 0x80;
				}
				else
				{
					s_nState = 0;
					break;
				}
				if (s_nBit < 16)
				{
					// カスタムコード
					s_wCst >>= 1;
					s_wCst |= (UINT16)b << 8;
				}
				else if (s_nBit < 24)
				{
					// データコード
					s_bDat >>= 1;
					s_bDat |= b;
				}
				else if (s_nBit < 32)
				{
					// データコード(反転)
					s_nDat >>= 1;
					s_nDat |= b;
				}
				if (++s_nBit == 32)
				{
					// 終了
					if (s_wCst == CUSTOM_CODE)
					{
						if ((s_bDat & s_nDat) == 0) // Check
						{
							s_nRem = ToKey(s_bDat);
							nKey = s_nRem;
						}
						else {
							s_nRem = KEY_NONE;
						}
						s_nRep = 0;
					}
					s_nState = 0;
				}
				else
				{
					s_nStart = IR_TMR;
					s_nState--;
				}
			}
			else
			{
				if (GetElapsed2(s_nStart) > HIDAT_L) {
					s_nState = 0;
				}
			}
			break;
		}
	}
	while (s_nState != 0 && nKey == KEY_NONE);

	return nKey;
}

// ●リモコンコードからキーコードへ変換
static UINT8 ToKey(UINT8 code)
{
	switch (code)
	{
	case 0x01: return KEY_ENTER;
	case 0x02: return KEY_BTN1;
	case 0x03: return KEY_BTN2;
	case 0x05: return KEY_VUP;
	case 0x04: return KEY_VDN;
	}
	return KEY_NONE;
}

//---------------------------------------------------------------

// ●I2C送信(I2C2)
static void SendI2C2(UINT8 nAddr, const UINT8* pData, int nCount)
{
	// Start conditon
	I2C2CONbits.SEN = 1;
	while (I2C2CONbits.SEN);

	// 7bit address & W flag
	I2C2TRN = (nAddr << 1) | 0x00;
	while (I2C2STATbits.TRSTAT);

	// data
	int i;
	for(i = 0; i < nCount; i++)
	{
		I2C2TRN = pData[i];
		while (I2C2STATbits.TRSTAT);
	}

	// Stop condition
	I2C2CONbits.PEN = 1;
	while (I2C2CONbits.PEN);
}

// ●I2C受信(I2C2)
static void RecvI2C2(UINT8 nAddr, UINT8* pData, int nCount)
{
	// Start conditon
	I2C2CONbits.SEN = 1;
	while (I2C2CONbits.SEN);

	// 7bit address & R flag
	I2C2TRN = (nAddr << 1) | 0x01;
	while (I2C2STATbits.TRSTAT);

	// data
	int i = 0;
	do {
		// data
		I2C2CONbits.RCEN = 1;
		while(I2C2CONbits.RCEN);
		pData[i++] = I2C2RCV;

		// ACK or NAK
		I2C2CONbits.ACKDT = (i == nCount)? 1 : 0;
		I2C2CONbits.ACKEN = 1;
		while(I2C2CONbits.ACKEN);

	} while(!I2C2CONbits.ACKDT);

	// Stop condition
	I2C2CONbits.PEN = 1;
	while (I2C2CONbits.PEN);
}

// ●オリジナルプロトコル送信
static void SendComm(const UINT8* pData, int nCount)
{
	const UINT8* p = pData;

	// Busy解除待ち
	while(!COM_CLK);

	COM_CLK_TRIS = 0;
	Nop();
	
	int i;
	for(i = 0; i < nCount; i++, p++)
	{
		UINT8 data = *p;
		
		COM_CLK_TRIS = 0;
		COM_DAT_TRIS = data >> 0;
		COM_CLK_TRIS = 1;
		Nop();
		COM_CLK_TRIS = 0;
		COM_DAT_TRIS = data >> 1;
		COM_CLK_TRIS = 1;
		Nop();
		COM_CLK_TRIS = 0;
		COM_DAT_TRIS = data >> 2;
		COM_CLK_TRIS = 1;
		Nop();
		COM_CLK_TRIS = 0;
		COM_DAT_TRIS = data >> 3;
		COM_CLK_TRIS = 1;
		Nop();
		COM_CLK_TRIS = 0;
		COM_DAT_TRIS = data >> 4;
		COM_CLK_TRIS = 1;
		Nop();
		COM_CLK_TRIS = 0;
		COM_DAT_TRIS = data >> 5;
		COM_CLK_TRIS = 1;
		Nop();
		COM_CLK_TRIS = 0;
		COM_DAT_TRIS = data >> 6;
		COM_CLK_TRIS = 1;
		Nop();
		COM_CLK_TRIS = 0;
		COM_DAT_TRIS = data >> 7;
		COM_CLK_TRIS = 1;
	}
}

// ●UART送信
static void SendUART(const char* pLine)
{
	// 1行テキスト送信
	const char* p = pLine;
	while(*p != '\0')
	{
		while(U1STAbits.UTXBF);
		U1TXREGbits.U1TXREG = *p;
		p++;
	}

	// 改行送信
	while(U1STAbits.UTXBF);
	U1TXREGbits.U1TXREG = '\r';
	while(U1STAbits.UTXBF);
	U1TXREGbits.U1TXREG = '\n';
}

// ●経過時間を取得する(1ms単位)
static UINT16 GetElapsed(UINT16 nBase)
{
	UINT16 nTick = s_nTick;
	if (nTick >= nBase)
		return (nTick - nBase);

	UINT16 nRet = 65535U - (nBase - nTick);
	return nRet + 1;
}

// ●経過時間を取得する(2us単位)
static UINT16 GetElapsed2(UINT16 nBase)
{
	UINT16 nTick = TMR4;
	if (nTick >= nBase)
		return (nTick - nBase);

	UINT16 nRet = 65535U - (nBase - nTick);
	return nRet + 1;
}

// ●MSウェイト
static void DelayMs(UINT16 ms)
{
	UINT16 nBase = s_nTick;
	for(;;)
	{
		if (GetElapsed(nBase) >= ms)
			return;

		ClrWdt();
	}
}

//---------------------------------------------------------------

// ●状態変化通知割り込み
void __attribute__((interrupt, shadow, auto_psv)) _CNInterrupt(void)
{
	// RE1_B
	BOOL bCur1B = RE1_B;
	BOOL bCur1A = RE1_A;
	if (bCur1A != s_bRE1A)
	{
		s_bRE1A = bCur1A;

		if (bCur1A) {
			s_nRE1 = (bCur1B)? RE_DOWN : RE_UP;
		}
		else {
			s_nRE1 = (bCur1B)? RE_UP : RE_DOWN;
		}
	}
	
	// RE2_A
	BOOL bCur2B = RE2_B;
	BOOL bCur2A = RE2_A;
	if (bCur2A != s_bRE2A)
	{
		s_bRE2A = bCur2A;

		if (bCur2A) {
			s_nRE2 = (bCur2B)? RE_DOWN : RE_UP;
		}
		else {
			s_nRE2 = (bCur2B)? RE_UP : RE_DOWN;
		}
	}
	
	// RE1_E
	if (!RE1_E)
		s_bSW1 = TRUE;
	
	// RE2_E
	if (!RE2_E)
		s_bSW2 = TRUE;

	_CNIF = 0;
}

// ●Timer1割り込み
void __attribute__((interrupt, auto_psv)) _T1Interrupt(void)
{
	// 1ms単位
	s_nTick++;
	_T1IF = 0;
}

// ●Timer2割り込み
void __attribute__((interrupt, auto_psv)) _T2Interrupt(void)
{
	SOUNDER = !SOUNDER;
	if (--s_nBeep <= 0)
	{
		T2CONbits.TON = 0;
		SOUNDER = 0;
	}
	_T2IF = 0;
}

// ●Timer3割り込み
void __attribute__((interrupt, auto_psv)) _T3Interrupt(void)
{
	DEBUG_LED = !DEBUG_LED;
	_T3IF = 0;
}

// ●Timer5割り込み
void __attribute__((interrupt, auto_psv)) _T5Interrupt(void)
{
	POWER_LED = !POWER_LED;
	_T5IF = 0;
}

// ●UART割り込み
void __attribute__((interrupt, auto_psv)) _U1RXInterrupt(void)
{
	static volatile int s_nCount = 0;
	
	char c =  U1RXREGbits.U1RXREG;
	if (c != '\r' && c !='\n')
		s_Uart[s_nCount++] = c;

	// 1行読み込み
	if (c == '\n' || (s_nCount >= sizeof(s_Uart) - 1))
	{
		s_Uart[s_nCount] = '\0';
		s_nCount = 0;
		
		s_bRecved = TRUE;
	}

	_U1RXIF = 0;
}
